/*  Pcsx2 - Pc Ps2 Emulator
 *  Copyright (C) 2002-2003  Pcsx2 Team
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdlib.h>
#include <string.h>

#include "Common.h"
#include "InterTables.h"
#include "ix86/ix86.h"
#include "iR5900.h"
#include "iMMI.h"
#include "iFPU.h"
#include "iCP0.h"
#include "VU0.h"
#include "VUmicro.h"
#include "VUflags.h"
#include "iVUmicro.h"
#include "iVUops.h"

#ifdef __MSCW32__
#pragma warning(disable:4244)
#pragma warning(disable:4761)
#endif

//Lower/Upper instructions can use that..
#define _Ft_ (( VU->code >> 16) & 0x1F)  // The rt part of the instruction register 
#define _Fs_ (( VU->code >> 11) & 0x1F)  // The rd part of the instruction register 
#define _Fd_ (( VU->code >>  6) & 0x1F)  // The sa part of the instruction register

#define _X (( VU->code>>24) & 0x1)
#define _Y (( VU->code>>23) & 0x1)
#define _Z (( VU->code>>22) & 0x1)
#define _W (( VU->code>>21) & 0x1)

#define _Fsf_ (( VU->code >> 21) & 0x03)
#define _Ftf_ (( VU->code >> 23) & 0x03)

#define _Imm11_ 	(s32)(VU->code & 0x400 ? 0xfffffc00 | (VU->code & 0x3ff) : VU->code & 0x3ff)
#define _UImm11_	(s32)(VU->code & 0x7ff)

#define VU_VFx_ADDR(x)  (u32)&VU->VF[x].UL[0]
#define VU_VFy_ADDR(x)  (u32)&VU->VF[x].UL[1]
#define VU_VFz_ADDR(x)  (u32)&VU->VF[x].UL[2]
#define VU_VFw_ADDR(x)  (u32)&VU->VF[x].UL[3]

#define VU_REGR_ADDR    (u32)&VU->VI[REG_R]
#define VU_REGI_ADDR    (u32)&VU->VI[REG_I]
#define VU_REGQ_ADDR    (u32)&VU->VI[REG_Q]
#define VU_REGMAC_ADDR  (u32)&VU->VI[REG_MAC_FLAG]

#define VU_VI_ADDR(x)   (u32)&VU->VI[x].UL

#define VU_ACCx_ADDR    (u32)&VU->ACC.UL[0]
#define VU_ACCy_ADDR    (u32)&VU->ACC.UL[1]
#define VU_ACCz_ADDR    (u32)&VU->ACC.UL[2]
#define VU_ACCw_ADDR    (u32)&VU->ACC.UL[3]


#define float_to_int4(x)   (s32)((double)x / 0.0625f)
#define float_to_int12(x)  (s32)((double)x / 0.000244140625f)
#define float_to_int15(x)  (s32)((double)x / 0.000030517578125)

#define int4_to_float(x)   (float)((double)x * 0.0625f)
#define int12_to_float(x)  (float)((double)x * 0.000244140625f)
#define int15_to_float(x)  (float)((double)x * 0.000030517578125)

static VECTOR RDzero;

/********************************/
/*     HELP FUNCTIONS           */
/********************************/

#define _X_OR_Y   ((( VU->code >> 24 ) & 0x1 ) || (( VU->code >> 23 ) & 0x1 ) )
#define _Z_OR_W   ((( VU->code >> 22 ) & 0x1 ) || (( VU->code >> 21 ) & 0x1 ) )
#define _X_AND_Y  ((( VU->code >> 24 ) & 0x1 ) && (( VU->code >> 23 ) & 0x1 ) )
#define _Z_AND_W  ((( VU->code >> 22 ) & 0x1 ) && (( VU->code >> 21 ) & 0x1 ) )
#define _X_AND_Z  ((( VU->code >> 24 ) & 0x1 ) && (( VU->code >> 22 ) & 0x1 ) )
#define _Y_AND_Z  ((( VU->code >> 23 ) & 0x1 ) && (( VU->code >> 22 ) & 0x1 ) )
#define _Y_AND_W  ((( VU->code >> 23 ) & 0x1 ) && (( VU->code >> 21 ) & 0x1 ) )
#define _X_Y_Z_W  ((( VU->code >> 21 ) & 0xF ) )

// Variable to keep the last updated float vu reg
u32      vuLastUpdatedRegs[ 4 ] = { 0x0, 0x0, 0x0, 0x0 };
u32      vuLastUpdatedFields = 0x0;
VURegs   *vuLastVU = NULL;

float recMult_float_to_int4[4] = { (1.0f / 0.0625f), (1.0f / 0.0625f), (1.0f / 0.0625f), (1.0f / 0.0625f) };
float recMult_float_to_int12[4] = { (1.0f / 0.000244140625f), (1.0f / 0.000244140625f), (1.0f / 0.000244140625f), (1.0f / 0.000244140625f) };
float recMult_float_to_int15[4] = { (1.0f / 0.000030517578125), (1.0f / 0.000030517578125), (1.0f / 0.000030517578125), (1.0f / 0.000030517578125) };

void SSE_MOV_VFtoR( VURegs *VU, int fs )
{
    MOVAPSMtoR( XMM0, VU_VFx_ADDR( fs ) ); 
}

void SSE_MOV_ACCtoR( VURegs *VU )
{
   MOVAPSMtoR( XMM0, VU_ACCx_ADDR ); 
}

#ifdef __MSCW32__
__declspec(align(16)) 
#endif
u32 SSEmovMask[ 16 ][ 4 ] =
{
{ 0x00000000, 0x00000000, 0x00000000, 0x00000000 },
{ 0x00000000, 0x00000000, 0x00000000, 0xFFFFFFFF },
{ 0x00000000, 0x00000000, 0xFFFFFFFF, 0x00000000 },
{ 0x00000000, 0x00000000, 0xFFFFFFFF, 0xFFFFFFFF },
{ 0x00000000, 0xFFFFFFFF, 0x00000000, 0x00000000 },
{ 0x00000000, 0xFFFFFFFF, 0x00000000, 0xFFFFFFFF },
{ 0x00000000, 0xFFFFFFFF, 0xFFFFFFFF, 0x00000000 },
{ 0x00000000, 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF },
{ 0xFFFFFFFF, 0x00000000, 0x00000000, 0x00000000 },
{ 0xFFFFFFFF, 0x00000000, 0x00000000, 0xFFFFFFFF },
{ 0xFFFFFFFF, 0x00000000, 0xFFFFFFFF, 0x00000000 },
{ 0xFFFFFFFF, 0x00000000, 0xFFFFFFFF, 0xFFFFFFFF },
{ 0xFFFFFFFF, 0xFFFFFFFF, 0x00000000, 0x00000000 },
{ 0xFFFFFFFF, 0xFFFFFFFF, 0x00000000, 0xFFFFFFFF },
{ 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF, 0x00000000 },
{ 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF }
};

void _unpackVF_xyzw(int dstreg, int srcreg, int xyzw) {
	switch (xyzw) {
		case 0:
			MOVSSRtoR(dstreg, srcreg); 
			UNPCKLPSRtoR(dstreg, dstreg);
			MOVLHPSRtoR(dstreg, dstreg);
			break;
		case 1:
			MOVAPSRtoR(dstreg, srcreg); 
			UNPCKLPSRtoR(dstreg, dstreg);
			MOVHLPSRtoR(dstreg, dstreg);
			break;
		case 2:
			MOVAPSRtoR(dstreg, srcreg); 
			UNPCKHPSRtoR(dstreg, dstreg);
			MOVLHPSRtoR(dstreg, dstreg);
			break;
		case 3:
			MOVAPSRtoR(dstreg, srcreg); 
			UNPCKHPSRtoR(dstreg, dstreg);
			MOVHLPSRtoR(dstreg, dstreg);
			break;
	}
}

#define TYPE_TEMP	0
#define TYPE_VFREG	1
#define TYPE_ACC	2

typedef struct {
	int inuse;
	int type;
	int vfreg;
	int mode;
	int needed;
} _xmmregs;

#ifdef __x86_64__
#define XMMREGS 16
#else
#define XMMREGS 8
#endif

_xmmregs xmmregs[XMMREGS];

void _initXMMregs(VURegs *VU) {
	memset(xmmregs, 0, sizeof(xmmregs));
}

int  _getFreeXMMreg(VURegs *VU) {
	int i;

	for (i=0; i<XMMREGS; i++) {
		if (xmmregs[i].inuse == 0) return i;
	}
	for (i=0; i<XMMREGS; i++) {
		if (xmmregs[i].needed == 1) continue;

		_freeXMMreg(VU, i);
		return i;
	}
	SysPrintf("*PCSX2*: VUrec ERROR\n");

	return -1;
}

int _allocTempXMMreg(VURegs *VU, int xmmreg) {
	if (xmmreg == -1) {
		xmmreg = _getFreeXMMreg(VU);
	}

	xmmregs[xmmreg].inuse = 1;
	xmmregs[xmmreg].type = TYPE_TEMP;
	xmmregs[xmmreg].needed = 1;

//SysPrintf("_allocTempXMMreg => %d\n", xmmreg);
	return xmmreg;
}

int _allocVFtoXMMreg(VURegs *VU, int xmmreg, int vfreg, int mode) {
	int i;

	for (i=0; i<XMMREGS; i++) {
		if (xmmregs[i].inuse == 0) continue;
		if (xmmregs[i].type != TYPE_VFREG) continue;
		if (xmmregs[i].vfreg != vfreg) continue;

		xmmregs[i].mode|= mode;
//SysPrintf("_allocVFtoXMMreg => %d (inuse)\n", i);
		return i;
	}

	if (xmmreg == -1) {
		xmmreg = _getFreeXMMreg(VU);
	}

	xmmregs[xmmreg].inuse = 1;
	xmmregs[xmmreg].type = TYPE_VFREG;
	xmmregs[xmmreg].vfreg = vfreg;
	xmmregs[xmmreg].mode = mode;
	xmmregs[xmmreg].needed = 1;
	if (mode & MODE_READ) {
		MOVAPSMtoR(xmmreg, VU_VFx_ADDR(xmmregs[xmmreg].vfreg));
	}

//SysPrintf("_allocVFtoXMMreg => %d\n", xmmreg);
	return xmmreg;
}

int _allocACCtoXMMreg(VURegs *VU, int xmmreg, int mode) {
	int i;

	for (i=0; i<XMMREGS; i++) {
		if (xmmregs[i].inuse == 0) continue;
		if (xmmregs[i].type != TYPE_ACC) continue;

		xmmregs[i].mode|= mode;
		return i;
	}

	if (xmmreg == -1) {
		xmmreg = _getFreeXMMreg(VU);
	}

	xmmregs[xmmreg].inuse = 1;
	xmmregs[xmmreg].type = TYPE_ACC;
	xmmregs[xmmreg].mode = mode;
	xmmregs[xmmreg].needed = 1;
	if (mode & MODE_READ) {
		MOVAPSMtoR(xmmreg, VU_ACCx_ADDR);
	}

	return xmmreg;
}

void _addNeededVFtoXMMreg(VURegs *VU, int vfreg) {
	int i;

	for (i=0; i<XMMREGS; i++) {
		if (xmmregs[i].inuse == 0) continue;
		if (xmmregs[i].type != TYPE_VFREG) continue;
		if (xmmregs[i].vfreg != vfreg) continue;

		xmmregs[i].needed = 1;
	}
}

void _addNeededACCtoXMMreg(VURegs *VU) {
	int i;

	for (i=0; i<XMMREGS; i++) {
		if (xmmregs[i].inuse == 0) continue;
		if (xmmregs[i].type != TYPE_ACC) continue;

		xmmregs[i].needed = 1;
	}
}

void _clearNeededXMMregs(VURegs *VU) {
	int i;

	for (i=0; i<XMMREGS; i++) {
		xmmregs[i].needed = 0;
	}
}

void _freeXMMreg(VURegs *VU, int xmmreg) {
	if (xmmregs[xmmreg].type == TYPE_VFREG &&
		xmmregs[xmmreg].mode & MODE_WRITE) {
		MOVAPSRtoM(VU_VFx_ADDR(xmmregs[xmmreg].vfreg), xmmreg);
	}
	if (xmmregs[xmmreg].type == TYPE_ACC &&
		xmmregs[xmmreg].mode & MODE_WRITE) {
		MOVAPSRtoM(VU_ACCx_ADDR, xmmreg);
	}
	xmmregs[xmmreg].inuse = 0;
}

void _freeXMMregs(VURegs *VU) {
	int i;

//SysPrintf("_freeXMMregs\n");
	for (i=0; i<XMMREGS; i++) {
		if (xmmregs[i].inuse == 0) continue;

		_freeXMMreg(VU, i);
	}
}




void _recvuFMACflush(VURegs * VU) {
	int i;

	for (i=0; i<8; i++) {
		if (VU->fmac[i].enable == 0) continue;

		if ((vucycle - VU->fmac[i].sCycle) >= VU->fmac[i].Cycle) {
#ifdef VUM_LOG
//			if (Log) { VUM_LOG("flushing FMAC pipe[%d]\n", i); }
#endif
			VU->fmac[i].enable = 0;
			if (VU->fmac[i].reg == -REG_CLIP_FLAG) {
				MOV32MtoR(EAX, (u32)&VU->clipflag);
				MOV32RtoM((u32)&VU->VI[REG_CLIP_FLAG].UL, EAX);
//				SysPrintf("REG_CLIP_FLAG flush\n");
			}
		}
	}
}

void _recvuFDIVflush(VURegs * VU) {
	if (VU->fdiv.enable == 0) return;

	if ((vucycle - VU->fdiv.sCycle) >= VU->fdiv.Cycle) {
//		SysPrintf("flushing FDIV pipe\n");
		VU->fdiv.enable = 0;
		MOV32MtoR(EAX, (u32)&VU->q);
		MOV32RtoM((u32)&VU->VI[REG_Q].UL, EAX);
	}
}

void _recvuTestPipes(VURegs * VU) {
	_recvuFMACflush(VU);
	_recvuFDIVflush(VU);
}

void _recvuFMACTestStall(VURegs * VU, int reg) {
	int cycle;
	int i;

	for (i=0; i<8; i++) {
		if (VU->fmac[i].enable == 0) continue;
		if (VU->fmac[i].reg == reg) break;
	}

	if (i == 8) return;

	cycle = VU->fmac[i].Cycle - (vucycle - VU->fmac[i].sCycle);
	VU->fmac[i].enable = 0;
//	SysPrintf("FMAC stall %d\n", cycle);
	vucycle+= cycle;
	_recvuTestPipes(VU);
}

void _recvuFMACAdd(VURegs * VU, int reg) {
	int i;

	/* find a free fmac pipe */
	for (i=0; i<8; i++) {
		if (VU->fmac[i].enable == 1) continue;
		break;
	}
	if (i==8) {
		SysPrintf("*PCSX2*: error , out of fmacs\n");
	}

#ifdef VUM_LOG
//	if (Log) { VUM_LOG("adding FMAC pipe[%d]; reg %d\n", i, reg); }
#endif
	VU->fmac[i].enable = 1;
	VU->fmac[i].sCycle = vucycle;
	VU->fmac[i].Cycle = 3;
	VU->fmac[i].reg = reg; 
}

void _recvuFDIVAdd(VURegs * VU, int cycles) {
//	SysPrintf("adding FDIV pipe\n");
	VU->fdiv.enable = 1;
	VU->fdiv.sCycle = vucycle;
	VU->fdiv.Cycle  = cycles;
}

void _recvuTestFMACStalls(VURegs * VU, _VURegsNum *VUregsn) {
	if (VUregsn->VFread0) {
		_recvuFMACTestStall(VU, VUregsn->VFread0);
	}
	if (VUregsn->VFread1) {
		_recvuFMACTestStall(VU, VUregsn->VFread1);
	}
}

void _recvuAddFMACStalls(VURegs * VU, _VURegsNum *VUregsn) {
	if (VUregsn->VFwrite) {
		_recvuFMACAdd(VU, VUregsn->VFwrite);
	}

	if (VUregsn->VIwrite & (1 << REG_CLIP_FLAG)) {
//		SysPrintf("REG_CLIP_FLAG pipe\n");
		_recvuFMACAdd(VU, -REG_CLIP_FLAG);
	}
}

void _recvuFlushFDIV(VURegs * VU) {
	int cycle;

	if (VU->fdiv.enable == 0) return;

	cycle = VU->fdiv.Cycle - (vucycle - VU->fdiv.sCycle);
//	SysPrintf("waiting FDIV pipe %d\n", cycle);
	VU->fdiv.enable = 0;
	vucycle+= cycle;
	MOV32MtoR(EAX, (u32)&VU->q);
	MOV32RtoM((u32)&VU->VI[REG_Q].UL, EAX);
}

void _recvuTestFDIVStalls(VURegs * VU, _VURegsNum *VUregsn) {
//	_vuTestFMACStalls(VURegs * VU, _VURegsNum *VUregsn);
	_recvuFlushFDIV(VU);
}

void _recvuAddFDIVStalls(VURegs * VU, _VURegsNum *VUregsn) {
//	_vuTestFMACStalls(VURegs * VU, _VURegsNum *VUregsn);
	if (VUregsn->VIwrite & (1 << REG_Q)) {
		_recvuFDIVAdd(VU, VU->qcycles);
	}
}

void _recvuTestUpperStalls(VURegs * VU, _VURegsNum *VUregsn) {
	switch (VUregsn->pipe) {
		case VUPIPE_FMAC: _recvuTestFMACStalls(VU, VUregsn); break;
	}
}

void _recvuTestLowerStalls(VURegs * VU, _VURegsNum *VUregsn) {
	switch (VUregsn->pipe) {
		case VUPIPE_FMAC: _recvuTestFMACStalls(VU, VUregsn); break;
		case VUPIPE_FDIV: _recvuTestFDIVStalls(VU, VUregsn); break;
	}
}

void _recvuAddUpperStalls(VURegs * VU, _VURegsNum *VUregsn) {
	switch (VUregsn->pipe) {
		case VUPIPE_FMAC: _recvuAddFMACStalls(VU, VUregsn); break;
	}
}

void _recvuAddLowerStalls(VURegs * VU, _VURegsNum *VUregsn) {
	switch (VUregsn->pipe) {
		case VUPIPE_FMAC: _recvuAddFMACStalls(VU, VUregsn); break;
		case VUPIPE_FDIV: _recvuAddFDIVStalls(VU, VUregsn); break;
	}
}

#if 0

void _vurecAnalyzeOp(VURegs *VU, _vuopinfo *info) {
	_VURegsNum lregs;
	_VURegsNum uregs;
	u32 *ptr; 
	int vfreg;
	int vireg;
	int i;

	ptr = (u32*)&VU1.Micro[ pc ]; 
	pc += 8; 

	if (ptr[1] & 0x40000000) { 
		branch = 2; 
	} 
 
	VU1.code = ptr[1]; 
	VU1regs_UPPER_OPCODE[VU1.code & 0x3f](&uregs);

	/* check upper flags */ 
	if (ptr[1] & 0x80000000) { /* I flag */ 
		VU1.code = ptr[1]; 
	} else {
		VU1.code = ptr[0]; 
		VU1regs_LOWER_OPCODE[VU1.code >> 25](&lregs);

		if (uregs.VIwrite & (1 << REG_CLIP_FLAG)) {
			info->clip = 2;
		}

		if (uregs.VIread & (1 << REG_CLIP_FLAG)) {
			info->clip = 1;
		}

		VU1.code = ptr[1]; 

		VU1.code = ptr[0]; 
	}
}

void _vurecAnalyzeOp(VURegs *VU, _vuopinfo *info, int count) {
	int i, j;

	for (i=0; i<count; i++) {
		if (info[i].clip & 2) {
			for (j=i; j<count; j++) {
				if (info[j].clip) break;
			}
			if (j == count) {
			} else {
			}
		}
	}
}
#endif

/******************************/
/*        VU Flags            */
/******************************/

#define     COP2_FLOAT_MAX 2147483648.0
#define     COP2_FLOAT_MIN -2147483648.0

void recResetFlags( void )
{
   vuLastUpdatedRegs[ 0 ] = 0x0;
   vuLastUpdatedRegs[ 1 ] = 0x0;
   vuLastUpdatedRegs[ 2 ] = 0x0;
   vuLastUpdatedRegs[ 3 ] = 0x0;
   vuLastUpdatedFields = 0x0;
   vuLastVU = NULL;
}

//void recVUMI_STAT_UPDATE( VURegs * VU )
//{
//   if ( VU->VI[REG_MAC_FLAG].UL & 0x000f )
//   {
//      VU->VI[REG_STATUS_FLAG].UL |= 0x1;
//   }
//   if ( VU->VI[REG_MAC_FLAG].UL & 0x00f0 )
//   {
//      VU->VI[REG_STATUS_FLAG].UL |= 0x2;
//   }
//   if ( VU->VI[REG_MAC_FLAG].UL & 0x0f00 )
//   {
//      VU->VI[REG_STATUS_FLAG].UL |= 0x4;
//   }
//   if ( VU->VI[REG_MAC_FLAG].UL & 0xf000 )
//   {
//      VU->VI[REG_STATUS_FLAG].UL |= 0x8;
//   }
//   VU->VI[REG_STATUS_FLAG].UL |= ( VU->VI[REG_STATUS_FLAG].UL & 0xf ) << 6;
//}

void recVuUpdateDI( VURegs * VU )
{
	u32 Flag_S = 0;
	u32 Flag_I = 0;
	u32 Flag_D = 0;

	/*
	FLAG D - I
	*/
	if ( ( VU->VI[REG_MAC_FLAG].UL & 0x00F0 ) != 0 ) 
   {
      Flag_S = 1;
   }
	Flag_I = ( VU->VI[REG_STATUS_FLAG].UL >> 4 ) & 0x1;
	Flag_D = ( VU->VI[REG_STATUS_FLAG].UL >> 5 ) & 0x1;

	VU->VI[REG_STATUS_FLAG].UL |= ( Flag_I | Flag_S ) << 10;
	VU->VI[REG_STATUS_FLAG].UL |= ( Flag_D | Flag_S ) << 11;
}

void recVU_UPDATE_STAT_REG_DI0( VURegs * VU ) 
{
#if __x86_64__
   MOV32ItoR( EDI, (u32)VU );
   CALLFunc( (u32)recVuUpdateDI );
#else
   PUSH32I( (u32)VU );
   CALLFunc( (u32)recVuUpdateDI );
   ADD32ItoR( ESP, 1 * 4 );
#endif
}

void recVU_UPDATE_STAT_REG_DI1( VURegs * VU )
{
   OR32ItoM( (u32)&VU->VI[REG_STATUS_FLAG].UL, 0x20 );
   AND32ItoM( (u32)&VU->VI[REG_STATUS_FLAG].UL, ~0x10 );

#if __x86_64__
   MOV32ItoR( EDI, (u32)VU );
   CALLFunc( (u32)recVuUpdateDI );
#else
   PUSH32I( (u32)VU );
   CALLFunc( (u32)recVuUpdateDI );
   ADD32ItoR( ESP, 1 * 4 );
#endif
}

void recVU_UPDATE_STAT_REG_DI2( VURegs * VU )
{
   OR32ItoM( (u32)&VU->VI[REG_STATUS_FLAG].UL, 0x10 );
   AND32ItoM( (u32)&VU->VI[REG_STATUS_FLAG].UL, ~0x20 );

#if __x86_64__
   MOV32ItoR( EDI, (u32)VU );
   CALLFunc( (u32)recVuUpdateDI );
#else
   PUSH32I( (u32)VU );
   CALLFunc( (u32)recVuUpdateDI );
   ADD32ItoR( ESP, 1 * 4 );
#endif
}

void recVU_UPDATE_STAT_REG_DI3( VURegs * VU )
{
   AND32ItoM( (u32)&VU->VI[REG_STATUS_FLAG].UL, ~0x30 );

#if __x86_64__
   MOV32ItoR( EDI, (u32)VU );
   CALLFunc( (u32)recVuUpdateDI );
#else
   PUSH32I( (u32)VU );
   CALLFunc( (u32)recVuUpdateDI );
   ADD32ItoR( ESP, 1 * 4 );
#endif
}

void recVUMI_MACx_UPDATE( VURegs *VU, VECTOR * reg )
{
   if ( reg->f.x == 0.f )  VU->VI[ REG_MAC_FLAG ].UL |=  0x0008; 
   else                    VU->VI[ REG_MAC_FLAG ].UL &= ~0x0008;
   if ( reg->f.x < 0.f )   VU->VI[ REG_MAC_FLAG ].UL |=  0x0080; 
   else                    VU->VI[ REG_MAC_FLAG ].UL &= ~0x0080;
   if ( reg->f.x < COP2_FLOAT_MIN ) VU->VI[ REG_MAC_FLAG ].UL |=  0x0800;
   else                             VU->VI[ REG_MAC_FLAG ].UL &= ~0x0800;
   if ( reg->f.x > COP2_FLOAT_MAX ) VU->VI[ REG_MAC_FLAG ].UL |=  0x8000;
   else                             VU->VI[ REG_MAC_FLAG ].UL &= ~0x8000;
}

void recVUMI_MACy_UPDATE( VURegs *VU, VECTOR * reg )
{
   if ( reg->f.y == 0.f )  VU->VI[ REG_MAC_FLAG ].UL |=  0x0004; 
   else                    VU->VI[ REG_MAC_FLAG ].UL &= ~0x0004;
   if ( reg->f.y < 0.f )   VU->VI[ REG_MAC_FLAG ].UL |=  0x0040; 
   else                    VU->VI[ REG_MAC_FLAG ].UL &= ~0x0040;
   if ( reg->f.y < COP2_FLOAT_MIN ) VU->VI[ REG_MAC_FLAG ].UL |=  0x0400;
   else                             VU->VI[ REG_MAC_FLAG ].UL &= ~0x0400;
   if ( reg->f.y > COP2_FLOAT_MAX ) VU->VI[ REG_MAC_FLAG ].UL |=  0x4000;
   else                             VU->VI[ REG_MAC_FLAG ].UL &= ~0x4000;
}

void recVUMI_MACz_UPDATE( VURegs *VU, VECTOR * reg )
{
   if ( reg->f.z == 0.f )  VU->VI[ REG_MAC_FLAG ].UL |=  0x0002; 
   else                    VU->VI[ REG_MAC_FLAG ].UL &= ~0x0002;
   if ( reg->f.z < 0.f )   VU->VI[ REG_MAC_FLAG ].UL |=  0x0020; 
   else                    VU->VI[ REG_MAC_FLAG ].UL &= ~0x0020;
   if ( reg->f.z < COP2_FLOAT_MIN ) VU->VI[ REG_MAC_FLAG ].UL |=  0x0200;
   else                             VU->VI[ REG_MAC_FLAG ].UL &= ~0x0200;
   if ( reg->f.z > COP2_FLOAT_MAX ) VU->VI[ REG_MAC_FLAG ].UL |=  0x2000;
   else                             VU->VI[ REG_MAC_FLAG ].UL &= ~0x2000;
}

void recVUMI_MACw_UPDATE( VURegs *VU, VECTOR * reg )
{
   if ( reg->f.w == 0.f )  VU->VI[ REG_MAC_FLAG ].UL |=  0x0001; 
   else                    VU->VI[ REG_MAC_FLAG ].UL &= ~0x0001;
   if ( reg->f.w < 0.f )   VU->VI[ REG_MAC_FLAG ].UL |=  0x0010; 
   else                    VU->VI[ REG_MAC_FLAG ].UL &= ~0x0010;
   if ( reg->f.w < COP2_FLOAT_MIN ) VU->VI[ REG_MAC_FLAG ].UL |=  0x0100;
   else                             VU->VI[ REG_MAC_FLAG ].UL &= ~0x0100;
   if ( reg->f.w > COP2_FLOAT_MAX ) VU->VI[ REG_MAC_FLAG ].UL |=  0x1000;
   else                             VU->VI[ REG_MAC_FLAG ].UL &= ~0x1000;
}

void recVUMI_SetFlagsXYZW( VURegs *VU, u32 flags, float* x, float* y, float* z, float* w )
{
   if ( flags & 0x8 )
   {
      VU->VI[ REG_MAC_FLAG ].UL &= ~0x8888;
      if ( *x == 0.f )                    
      {
         VU->VI[ REG_MAC_FLAG ].UL |=  0x0008;
      }
      else 
      {
         if ( *x < 0.f )
         {
            VU->VI[ REG_MAC_FLAG ].UL |=  0x0080; 
            if ( *x < COP2_FLOAT_MIN )
            {
               VU->VI[ REG_MAC_FLAG ].UL |=  0x0800;
            }
         }
         else if ( *x > COP2_FLOAT_MAX )  
         {
            VU->VI[ REG_MAC_FLAG ].UL |=  0x8000;
         }
      }
      //if ( *x == 0.f )           VU->VI[ REG_MAC_FLAG ].UL |=  0x0008; 
      //else                       VU->VI[ REG_MAC_FLAG ].UL &= ~0x0008;
      //if ( *x < 0.f )            VU->VI[ REG_MAC_FLAG ].UL |=  0x0080; 
      //else                       VU->VI[ REG_MAC_FLAG ].UL &= ~0x0080;
      //if ( *x < COP2_FLOAT_MIN ) VU->VI[ REG_MAC_FLAG ].UL |=  0x0800;
      //else                       VU->VI[ REG_MAC_FLAG ].UL &= ~0x0800;
      //if ( *x > COP2_FLOAT_MAX ) VU->VI[ REG_MAC_FLAG ].UL |=  0x8000;
      //else                       VU->VI[ REG_MAC_FLAG ].UL &= ~0x8000;
   }

   if ( flags & 0x4 )
   {
      VU->VI[ REG_MAC_FLAG ].UL &= ~0x4444;
      if ( *y == 0.f )                    
      {
         VU->VI[ REG_MAC_FLAG ].UL |=  0x0004;
      }
      else 
      {
         if ( *y < 0.f )
         {
            VU->VI[ REG_MAC_FLAG ].UL |=  0x0040; 
            if ( *y < COP2_FLOAT_MIN )
            {
               VU->VI[ REG_MAC_FLAG ].UL |=  0x0400;
            }
         }
         else if ( *y > COP2_FLOAT_MAX )  
         {
            VU->VI[ REG_MAC_FLAG ].UL |=  0x4000;
         }
      }
      //if ( *y == 0.f )           VU->VI[ REG_MAC_FLAG ].UL |=  0x0004; 
      //else                       VU->VI[ REG_MAC_FLAG ].UL &= ~0x0004;
      //if ( *y < 0.f )            VU->VI[ REG_MAC_FLAG ].UL |=  0x0040; 
      //else                       VU->VI[ REG_MAC_FLAG ].UL &= ~0x0040;
      //if ( *y < COP2_FLOAT_MIN ) VU->VI[ REG_MAC_FLAG ].UL |=  0x0400;
      //else                       VU->VI[ REG_MAC_FLAG ].UL &= ~0x0400;
      //if ( *y > COP2_FLOAT_MAX ) VU->VI[ REG_MAC_FLAG ].UL |=  0x4000;
      //else                       VU->VI[ REG_MAC_FLAG ].UL &= ~0x4000;
   }

   if ( flags & 0x2 )
   {
      VU->VI[ REG_MAC_FLAG ].UL &= ~0x2222;
      if ( *z == 0.f )                    
      {
         VU->VI[ REG_MAC_FLAG ].UL |=  0x0002;
      }
      else 
      {
         if ( *z < 0.f )
         {
            VU->VI[ REG_MAC_FLAG ].UL |=  0x0020; 
            if ( *z < COP2_FLOAT_MIN )
            {
               VU->VI[ REG_MAC_FLAG ].UL |=  0x0200;
            }
         }
         else if ( *z > COP2_FLOAT_MAX )  
         {
            VU->VI[ REG_MAC_FLAG ].UL |=  0x2000;
         }
      }
      //if ( *z == 0.f )           VU->VI[ REG_MAC_FLAG ].UL |=  0x0002; 
      //else                       VU->VI[ REG_MAC_FLAG ].UL &= ~0x0002;
      //if ( *z < 0.f )            VU->VI[ REG_MAC_FLAG ].UL |=  0x0020; 
      //else                       VU->VI[ REG_MAC_FLAG ].UL &= ~0x0020;
      //if ( *z < COP2_FLOAT_MIN ) VU->VI[ REG_MAC_FLAG ].UL |=  0x0200;
      //else                       VU->VI[ REG_MAC_FLAG ].UL &= ~0x0200;
      //if ( *z > COP2_FLOAT_MAX ) VU->VI[ REG_MAC_FLAG ].UL |=  0x2000;
      //else                       VU->VI[ REG_MAC_FLAG ].UL &= ~0x2000;
   }

   if ( flags & 0x1 )
   {
      VU->VI[ REG_MAC_FLAG ].UL &= ~0x1111;
      if ( *w == 0.f )                    
      {
         VU->VI[ REG_MAC_FLAG ].UL |=  0x0001;
      }
      else 
      {
         if ( *w < 0.f )
         {
            VU->VI[ REG_MAC_FLAG ].UL |=  0x0010; 
            if ( *w < COP2_FLOAT_MIN )
            {
               VU->VI[ REG_MAC_FLAG ].UL |=  0x0100;
            }
         }
         else if ( *w > COP2_FLOAT_MAX )  
         {
            VU->VI[ REG_MAC_FLAG ].UL |=  0x1000;
         }
      }
      //if ( *w == 0.f )           VU->VI[ REG_MAC_FLAG ].UL |=  0x0001; 
      //else                       VU->VI[ REG_MAC_FLAG ].UL &= ~0x0001;
      //if ( *w < 0.f )            VU->VI[ REG_MAC_FLAG ].UL |=  0x0010; 
      //else                       VU->VI[ REG_MAC_FLAG ].UL &= ~0x0010;
      //if ( *w < COP2_FLOAT_MIN ) VU->VI[ REG_MAC_FLAG ].UL |=  0x0100;
      //else                       VU->VI[ REG_MAC_FLAG ].UL &= ~0x0100;
      //if ( *w > COP2_FLOAT_MAX ) VU->VI[ REG_MAC_FLAG ].UL |=  0x1000;
      //else                       VU->VI[ REG_MAC_FLAG ].UL &= ~0x1000;
   }

//   recVUMI_STAT_UPDATE( VU );
   if ( VU->VI[REG_MAC_FLAG].UL & 0x000f )
   {
      VU->VI[REG_STATUS_FLAG].UL |= 0x1;
   }
   if ( VU->VI[REG_MAC_FLAG].UL & 0x00f0 )
   {
      VU->VI[REG_STATUS_FLAG].UL |= 0x2;
   }
   if ( VU->VI[REG_MAC_FLAG].UL & 0x0f00 )
   {
      VU->VI[REG_STATUS_FLAG].UL |= 0x4;
   }
   if ( VU->VI[REG_MAC_FLAG].UL & 0xf000 )
   {
      VU->VI[REG_STATUS_FLAG].UL |= 0x8;
   }
   VU->VI[REG_STATUS_FLAG].UL |= ( VU->VI[REG_STATUS_FLAG].UL & 0xf ) << 6;
}

void recUpdateFlags( void )
{
   if ( ! vuLastVU )
   {
      return;
   }

   iFlushCall( );

   if ( vuLastUpdatedFields )
   {
#ifdef __x86_64__
      MOV32ItoR( R9, vuLastUpdatedRegs[ 3 ] );
      MOV32ItoR( R8, vuLastUpdatedRegs[ 2 ] );
      MOV32ItoR( ECX, vuLastUpdatedRegs[ 1 ] );
      MOV32ItoR( EDX, vuLastUpdatedRegs[ 0 ] );
      MOV32ItoR( ESI, vuLastUpdatedFields );
      MOV32ItoR( EDI, (u32)vuLastVU );
      CALLFunc( (u32)recVUMI_SetFlagsXYZW );
      vuLastUpdatedFields = 0;
#else
      PUSH32I( vuLastUpdatedRegs[ 3 ] );
      PUSH32I( vuLastUpdatedRegs[ 2 ] );
      PUSH32I( vuLastUpdatedRegs[ 1 ] );
      PUSH32I( vuLastUpdatedRegs[ 0 ] );
      PUSH32I( vuLastUpdatedFields );
      PUSH32I( (u32)vuLastVU );
      CALLFunc( (u32)recVUMI_SetFlagsXYZW );
      ADD32ItoR( ESP, 6 * 4 );
      vuLastUpdatedFields = 0;
#endif
   }
}

void recVUMI_STAT_REG_UPDATE( VURegs *VU, u32 dst )
{
   vuLastVU = VU;
   if ( _X )
   {
      vuLastUpdatedRegs[ 0 ] = dst;
   }
   if ( _Y )
   {
      vuLastUpdatedRegs[ 1 ] = dst + 4;
   }
   if ( _Z )
   {
      vuLastUpdatedRegs[ 2 ] = dst + 8;
   }
   if ( _W )
   {
      vuLastUpdatedRegs[ 3 ] = dst + 12;
   }
   vuLastUpdatedFields |= _X_Y_Z_W;
}


/******************************/
/*   VU Upper instructions    */
/******************************/

static u32 const_abs_table[16][4] = 
{
   { 0xffffffff, 0xffffffff, 0xffffffff, 0xffffffff },
   { 0xffffffff, 0xffffffff, 0xffffffff, 0x7fffffff },
   { 0xffffffff, 0xffffffff, 0x7fffffff, 0xffffffff },
   { 0xffffffff, 0xffffffff, 0x7fffffff, 0x7fffffff },
   { 0xffffffff, 0x7fffffff, 0xffffffff, 0xffffffff },
   { 0xffffffff, 0x7fffffff, 0xffffffff, 0x7fffffff },
   { 0xffffffff, 0x7fffffff, 0x7fffffff, 0xffffffff },
   { 0xffffffff, 0x7fffffff, 0x7fffffff, 0x7fffffff },
   { 0x7fffffff, 0xffffffff, 0xffffffff, 0xffffffff },
   { 0x7fffffff, 0xffffffff, 0xffffffff, 0x7fffffff },
   { 0x7fffffff, 0xffffffff, 0x7fffffff, 0xffffffff },
   { 0x7fffffff, 0xffffffff, 0x7fffffff, 0x7fffffff },
   { 0x7fffffff, 0x7fffffff, 0xffffffff, 0xffffffff },
   { 0x7fffffff, 0x7fffffff, 0xffffffff, 0x7fffffff },
   { 0x7fffffff, 0x7fffffff, 0x7fffffff, 0xffffffff },
   { 0x7fffffff, 0x7fffffff, 0x7fffffff, 0x7fffffff },
};

void recVUMI_ABS(VURegs *VU) {
    if ( _Ft_ == 0 ) return;

	_freeXMMregs(VU);
   CPU_SSE_START;

   MOVAPSMtoR( XMM0, VU_VFx_ADDR( _Fs_ ) );

   // Doing the abs
   ANDPSMtoR( XMM0, (u32)&const_abs_table[ _X_Y_Z_W ][ 0 ] );
   if ( _X_Y_Z_W != 0xf )
   {
      if ( _Ft_ != _Fs_ )
      {
         MOVAPSMtoR( XMM2, VU_VFx_ADDR( _Ft_ ) );
         ANDPSMtoR( XMM0, (u32)&SSEmovMask[ _X_Y_Z_W ][ 0 ] );
         ANDPSMtoR( XMM2, (u32)&SSEmovMask[ 15 - _X_Y_Z_W ][ 0 ] );
         ORPSRtoR( XMM0, XMM2 );
      }
   }

   MOVAPSRtoM( VU_VFx_ADDR( _Ft_ ), XMM0 );

   CPU_SSE_END;
}

void recVUMI_ADD(VURegs *VU) {
	int t0reg;
	int fsreg;
	int ftreg;
	int fdreg;

	_addNeededVFtoXMMreg(VU, _Fs_);
	_addNeededVFtoXMMreg(VU, _Ft_);
	if (_Fd_) _addNeededVFtoXMMreg(VU, _Fd_);
	fsreg = _allocVFtoXMMreg(VU, -1, _Fs_, MODE_READ);
	ftreg = _allocVFtoXMMreg(VU, -1, _Ft_, MODE_READ);

	if (_X_Y_Z_W != 0xf) {
		if (_Fd_) fdreg = _allocVFtoXMMreg(VU, -1, _Fd_, MODE_READ | MODE_WRITE);
		else fdreg = _allocTempXMMreg(VU, -1);
		t0reg = _allocTempXMMreg(VU, -1);

		MOVAPSRtoR(t0reg, fsreg);
		ADDPSRtoR(t0reg, ftreg);

		ANDPSMtoR(fdreg, (u32)&SSEmovMask[ 15 - _X_Y_Z_W ][ 0 ] );
		ANDPSMtoR(t0reg, (u32)&SSEmovMask[ _X_Y_Z_W ][ 0 ] );
		ORPSRtoR(fdreg, t0reg);

		_freeXMMreg(VU, t0reg);
	} else {
		if (_Fd_) fdreg = _allocVFtoXMMreg(VU, -1, _Fd_, MODE_WRITE);
		else fdreg = _allocTempXMMreg(VU, -1);

		MOVAPSRtoR(fdreg, fsreg);
		ADDPSRtoR(fdreg, ftreg);
	}

	if (_Fd_ == 0) _freeXMMreg(VU, fdreg);
	_clearNeededXMMregs(VU);

	recVUMI_STAT_REG_UPDATE( VU, VU_ACCx_ADDR );
}

void recVUMI_ADD_iq(VURegs *VU, u32 addr) {
	int t0reg;
	int fsreg;
	int fdreg;

	_addNeededVFtoXMMreg(VU, _Fs_);
	if (_Fd_) _addNeededVFtoXMMreg(VU, _Fd_);
	t0reg = _allocTempXMMreg(VU, -1);
	fsreg = _allocVFtoXMMreg(VU, -1, _Fs_, MODE_READ);

	MOVSSMtoR(t0reg, addr); 
	UNPCKLPSRtoR(t0reg, t0reg);
	MOVLHPSRtoR(t0reg, t0reg);

	if (_X_Y_Z_W != 0xf) {
		if (_Fd_) fdreg = _allocVFtoXMMreg(VU, -1, _Fd_, MODE_READ | MODE_WRITE);
		else fdreg = _allocTempXMMreg(VU, -1);

		ADDPSRtoR(t0reg, fsreg);

		ANDPSMtoR(fdreg, (u32)&SSEmovMask[ 15 - _X_Y_Z_W ][ 0 ] );
		ANDPSMtoR(t0reg, (u32)&SSEmovMask[ _X_Y_Z_W ][ 0 ] );
		ORPSRtoR(fdreg, t0reg);
	} else {
		if (_Fd_) fdreg = _allocVFtoXMMreg(VU, -1, _Fd_, MODE_WRITE);
		else fdreg = _allocTempXMMreg(VU, -1);

		MOVAPSRtoR(fdreg, fsreg);
		ADDPSRtoR(fdreg, t0reg);
	}

	_freeXMMreg(VU, t0reg);
	if (_Fd_ == 0) _freeXMMreg(VU, fdreg);
	_clearNeededXMMregs(VU);

	recVUMI_STAT_REG_UPDATE( VU, VU_ACCx_ADDR );
}

void recVUMI_ADD_xyzw(VURegs *VU, int xyzw) {
	int t0reg;
	int fsreg;
	int ftreg;
	int fdreg;

	_addNeededVFtoXMMreg(VU, _Fs_);
	_addNeededVFtoXMMreg(VU, _Ft_);
	if (_Fd_) _addNeededVFtoXMMreg(VU, _Fd_);
	t0reg = _allocTempXMMreg(VU, -1);
	fsreg = _allocVFtoXMMreg(VU, -1, _Fs_, MODE_READ);
	ftreg = _allocVFtoXMMreg(VU, -1, _Ft_, MODE_READ);

	_unpackVF_xyzw(t0reg, ftreg, xyzw);

	if (_X_Y_Z_W != 0xf) {
		if (_Fd_) fdreg = _allocVFtoXMMreg(VU, -1, _Fd_, MODE_READ | MODE_WRITE);
		else fdreg = _allocTempXMMreg(VU, -1);

		ADDPSRtoR(t0reg, fsreg);

		ANDPSMtoR(fdreg, (u32)&SSEmovMask[ 15 - _X_Y_Z_W ][ 0 ] );
		ANDPSMtoR(t0reg, (u32)&SSEmovMask[ _X_Y_Z_W ][ 0 ] );
		ORPSRtoR(fdreg, t0reg);
	} else {
		if (_Fd_) fdreg = _allocVFtoXMMreg(VU, -1, _Fd_, MODE_WRITE);
		else fdreg = _allocTempXMMreg(VU, -1);

		MOVAPSRtoR(fdreg, fsreg);
		ADDPSRtoR(fdreg, t0reg);
	}

	_freeXMMreg(VU, t0reg);
	if (_Fd_ == 0) _freeXMMreg(VU, fdreg);
	_clearNeededXMMregs(VU);

	recVUMI_STAT_REG_UPDATE( VU, VU_ACCx_ADDR );
}

void recVUMI_ADDi(VURegs *VU) { recVUMI_ADD_iq(VU, VU_REGI_ADDR); }
void recVUMI_ADDq(VURegs *VU) { recVUMI_ADD_iq(VU, VU_REGQ_ADDR); }
void recVUMI_ADDx(VURegs *VU) { recVUMI_ADD_xyzw(VU, 0); }
void recVUMI_ADDy(VURegs *VU) { recVUMI_ADD_xyzw(VU, 1); }
void recVUMI_ADDz(VURegs *VU) { recVUMI_ADD_xyzw(VU, 2); }
void recVUMI_ADDw(VURegs *VU) { recVUMI_ADD_xyzw(VU, 3); }

void recVUMI_ADDA(VURegs *VU) {
	int t0reg;
	int accreg;
	int fsreg;
	int ftreg;

	_addNeededVFtoXMMreg(VU, _Fs_);
	_addNeededVFtoXMMreg(VU, _Ft_);
	_addNeededACCtoXMMreg(VU);
	fsreg = _allocVFtoXMMreg(VU, -1, _Fs_, MODE_READ);
	ftreg = _allocVFtoXMMreg(VU, -1, _Ft_, MODE_READ);

	if (_X_Y_Z_W != 0xf) {
		t0reg = _allocTempXMMreg(VU, -1);
		accreg = _allocACCtoXMMreg(VU, -1, MODE_READ | MODE_WRITE);

		MOVAPSRtoR(t0reg, fsreg);
		ADDPSRtoR(t0reg, ftreg);

		ANDPSMtoR(accreg, (u32)&SSEmovMask[ 15 - _X_Y_Z_W ][ 0 ] );
		ANDPSMtoR(t0reg, (u32)&SSEmovMask[ _X_Y_Z_W ][ 0 ] );
		ORPSRtoR(accreg, t0reg);

		_freeXMMreg(VU, t0reg);
	} else {
		accreg = _allocACCtoXMMreg(VU, -1, MODE_WRITE);
		MOVAPSRtoR(accreg, fsreg);
		ADDPSRtoR(accreg, ftreg);
	}

	_clearNeededXMMregs(VU);
	recVUMI_STAT_REG_UPDATE( VU, VU_ACCx_ADDR );
}

void recVUMI_ADDA_iq(VURegs *VU, u32 addr) {
	int t0reg;
	int accreg;
	int fsreg;

	_addNeededVFtoXMMreg(VU, _Fs_);
	_addNeededACCtoXMMreg(VU);
	fsreg = _allocVFtoXMMreg(VU, -1, _Fs_, MODE_READ);

	if (_X_Y_Z_W != 0xf) {
		t0reg = _allocTempXMMreg(VU, -1);
		accreg = _allocACCtoXMMreg(VU, -1, MODE_READ | MODE_WRITE);

		MOVSSMtoR(t0reg, addr); 
		UNPCKLPSRtoR(t0reg, t0reg);
		MOVLHPSRtoR(t0reg, t0reg);

		ADDPSRtoR(t0reg, fsreg);

		ANDPSMtoR(accreg, (u32)&SSEmovMask[15 - _X_Y_Z_W][0]);
		ANDPSMtoR(t0reg, (u32)&SSEmovMask[_X_Y_Z_W][0]);
		ORPSRtoR(accreg, t0reg);

		_freeXMMreg(VU, t0reg);
	} else {
		accreg = _allocACCtoXMMreg(VU, -1, MODE_WRITE);

		MOVSSMtoR(accreg, addr); 
		UNPCKLPSRtoR(accreg, accreg);
		MOVLHPSRtoR(accreg, accreg);

		ADDPSRtoR(accreg, fsreg);
	}

	_clearNeededXMMregs(VU);
	recVUMI_STAT_REG_UPDATE(VU, VU_ACCx_ADDR);
}

void recVUMI_ADDAi(VURegs *VU) { recVUMI_ADDA_iq(VU, VU_REGI_ADDR); }
void recVUMI_ADDAq(VURegs *VU) { recVUMI_ADDA_iq(VU, VU_REGQ_ADDR); }

void recVUMI_ADDA_xyzw(VURegs *VU, int xyzw) {
	int t0reg;
	int accreg;
	int fsreg;
	int ftreg;

	_addNeededVFtoXMMreg(VU, _Fs_);
	_addNeededVFtoXMMreg(VU, _Ft_);
	_addNeededACCtoXMMreg(VU);
	t0reg = _allocTempXMMreg(VU, -1);
	fsreg = _allocVFtoXMMreg(VU, -1, _Fs_, MODE_READ);
	ftreg = _allocVFtoXMMreg(VU, -1, _Ft_, MODE_READ);

	_unpackVF_xyzw(t0reg, ftreg, xyzw);
	ADDPSRtoR(t0reg, fsreg);

	if (_X_Y_Z_W != 0xf) {
		accreg = _allocACCtoXMMreg(VU, -1, MODE_READ | MODE_WRITE);
		ANDPSMtoR(accreg, (u32)&SSEmovMask[15 - _X_Y_Z_W][0]);
		ANDPSMtoR(t0reg, (u32)&SSEmovMask[_X_Y_Z_W][0]);
		ORPSRtoR(accreg, t0reg);
	} else {
		accreg = _allocACCtoXMMreg(VU, -1, MODE_WRITE);
		MOVAPSRtoR(accreg, t0reg);
	}

	_clearNeededXMMregs(VU);
	recVUMI_STAT_REG_UPDATE(VU, VU_ACCx_ADDR);
}

void recVUMI_ADDAx(VURegs *VU) { recVUMI_ADDA_xyzw(VU, 0); }
void recVUMI_ADDAy(VURegs *VU) { recVUMI_ADDA_xyzw(VU, 1); }
void recVUMI_ADDAz(VURegs *VU) { recVUMI_ADDA_xyzw(VU, 2); }
void recVUMI_ADDAw(VURegs *VU) { recVUMI_ADDA_xyzw(VU, 3); }

void recVUMI_SUB(VURegs *VU) {
	int t0reg;
	int fsreg;
	int ftreg;
	int fdreg;

	_addNeededVFtoXMMreg(VU, _Fs_);
	_addNeededVFtoXMMreg(VU, _Ft_);
	if (_Fd_) _addNeededVFtoXMMreg(VU, _Fd_);
	fsreg = _allocVFtoXMMreg(VU, -1, _Fs_, MODE_READ);
	ftreg = _allocVFtoXMMreg(VU, -1, _Ft_, MODE_READ);

	if (_X_Y_Z_W != 0xf) {
		if (_Fd_) fdreg = _allocVFtoXMMreg(VU, -1, _Fd_, MODE_READ | MODE_WRITE);
		else fdreg = _allocTempXMMreg(VU, -1);
		t0reg = _allocTempXMMreg(VU, -1);

		MOVAPSRtoR(t0reg, fsreg);
		SUBPSRtoR(t0reg, ftreg);

		ANDPSMtoR(fdreg, (u32)&SSEmovMask[ 15 - _X_Y_Z_W ][ 0 ] );
		ANDPSMtoR(t0reg, (u32)&SSEmovMask[ _X_Y_Z_W ][ 0 ] );
		ORPSRtoR(fdreg, t0reg);

		_freeXMMreg(VU, t0reg);
	} else {
		if (_Fd_) fdreg = _allocVFtoXMMreg(VU, -1, _Fd_, MODE_WRITE);
		else fdreg = _allocTempXMMreg(VU, -1);

		MOVAPSRtoR(fdreg, fsreg);
		SUBPSRtoR(fdreg, ftreg);
	}

	if (_Fd_ == 0) _freeXMMreg(VU, fdreg);
	_clearNeededXMMregs(VU);

	recVUMI_STAT_REG_UPDATE( VU, VU_ACCx_ADDR );
}

void recVUMI_SUB_iq(VURegs *VU, u32 addr) {
	int t0reg;
	int t1reg;
	int fsreg;
	int fdreg;

	_addNeededVFtoXMMreg(VU, _Fs_);
	if (_Fd_) _addNeededVFtoXMMreg(VU, _Fd_);
	t0reg = _allocTempXMMreg(VU, -1);
	fsreg = _allocVFtoXMMreg(VU, -1, _Fs_, MODE_READ);

	MOVSSMtoR(t0reg, addr); 
	UNPCKLPSRtoR(t0reg, t0reg);
	MOVLHPSRtoR(t0reg, t0reg);

	if (_X_Y_Z_W != 0xf) {
		if (_Fd_) fdreg = _allocVFtoXMMreg(VU, -1, _Fd_, MODE_READ | MODE_WRITE);
		else fdreg = _allocTempXMMreg(VU, -1);
		t1reg = _allocTempXMMreg(VU, -1);

		MOVAPSRtoR(t1reg, fsreg);
		SUBPSRtoR(t1reg, t0reg);

		ANDPSMtoR(fdreg, (u32)&SSEmovMask[ 15 - _X_Y_Z_W ][ 0 ] );
		ANDPSMtoR(t1reg, (u32)&SSEmovMask[ _X_Y_Z_W ][ 0 ] );
		ORPSRtoR(fdreg, t1reg);

		_freeXMMreg(VU, t1reg);
	} else {
		if (_Fd_) fdreg = _allocVFtoXMMreg(VU, -1, _Fd_, MODE_WRITE);
		else fdreg = _allocTempXMMreg(VU, -1);

		MOVAPSRtoR(fdreg, fsreg);
		SUBPSRtoR(fdreg, t0reg);
	}

	_freeXMMreg(VU, t0reg);
	if (_Fd_ == 0) _freeXMMreg(VU, fdreg);
	_clearNeededXMMregs(VU);

	recVUMI_STAT_REG_UPDATE( VU, VU_ACCx_ADDR );
}

void recVUMI_SUB_xyzw(VURegs *VU, int xyzw) {
	int t0reg;
	int t1reg;
	int fsreg;
	int ftreg;
	int fdreg;

	_addNeededVFtoXMMreg(VU, _Fs_);
	_addNeededVFtoXMMreg(VU, _Ft_);
	if (_Fd_) _addNeededVFtoXMMreg(VU, _Fd_);
	t0reg = _allocTempXMMreg(VU, -1);
	fsreg = _allocVFtoXMMreg(VU, -1, _Fs_, MODE_READ);
	ftreg = _allocVFtoXMMreg(VU, -1, _Ft_, MODE_READ);

	_unpackVF_xyzw(t0reg, ftreg, xyzw);

	if (_X_Y_Z_W != 0xf) {
		if (_Fd_) fdreg = _allocVFtoXMMreg(VU, -1, _Fd_, MODE_READ | MODE_WRITE);
		else fdreg = _allocTempXMMreg(VU, -1);
		t1reg = _allocTempXMMreg(VU, -1);

		MOVAPSRtoR(t1reg, fsreg);
		SUBPSRtoR(t1reg, t0reg);

		ANDPSMtoR(fdreg, (u32)&SSEmovMask[ 15 - _X_Y_Z_W ][ 0 ] );
		ANDPSMtoR(t1reg, (u32)&SSEmovMask[ _X_Y_Z_W ][ 0 ] );
		ORPSRtoR(fdreg, t1reg);

		_freeXMMreg(VU, t1reg);
	} else {
		if (_Fd_) fdreg = _allocVFtoXMMreg(VU, -1, _Fd_, MODE_WRITE);
		else fdreg = _allocTempXMMreg(VU, -1);

		MOVAPSRtoR(fdreg, fsreg);
		SUBPSRtoR(fdreg, t0reg);
	}

	_freeXMMreg(VU, t0reg);
	if (_Fd_ == 0) _freeXMMreg(VU, fdreg);
	_clearNeededXMMregs(VU);

	recVUMI_STAT_REG_UPDATE( VU, VU_ACCx_ADDR );
}

void recVUMI_SUBi(VURegs *VU) { recVUMI_SUB_iq(VU, VU_REGI_ADDR); }
void recVUMI_SUBq(VURegs *VU) { recVUMI_SUB_iq(VU, VU_REGQ_ADDR); }
void recVUMI_SUBx(VURegs *VU) { recVUMI_SUB_xyzw(VU, 0); }
void recVUMI_SUBy(VURegs *VU) { recVUMI_SUB_xyzw(VU, 1); }
void recVUMI_SUBz(VURegs *VU) { recVUMI_SUB_xyzw(VU, 2); }
void recVUMI_SUBw(VURegs *VU) { recVUMI_SUB_xyzw(VU, 3); }

void recVUMI_SUBA( VURegs *VU )
{
	_freeXMMregs(VU);
   CPU_SSE_START;

   MOVAPSMtoR( XMM0, VU_VFx_ADDR( _Fs_ ) );
   SUBPSMtoR( XMM0, VU_VFx_ADDR( _Ft_ ) );

   if ( _X_Y_Z_W != 0xf )
   {
      MOVAPSMtoR( XMM2, VU_ACCx_ADDR );
      ANDPSMtoR( XMM0, (u32)&SSEmovMask[ _X_Y_Z_W ][ 0 ] );
      ANDPSMtoR( XMM2, (u32)&SSEmovMask[ 15 - _X_Y_Z_W ][ 0 ] );
      ORPSRtoR( XMM0, XMM2 );
   }

   MOVAPSRtoM( VU_ACCx_ADDR, XMM0 );

   recVUMI_STAT_REG_UPDATE( VU, VU_ACCx_ADDR );
   CPU_SSE_END;
}

void recVUMI_SUBA_iqxyzw( VURegs *VU, u32 regAddr ) 
{
	_freeXMMregs(VU);
   CPU_SSE_START;

   MOVAPSMtoR( XMM0, VU_VFx_ADDR( _Fs_ ) );
   MOVSSMtoR( XMM1, regAddr );
   UNPCKLPSRtoR( XMM1, XMM1 );
   MOVLHPSRtoR( XMM1, XMM1 );
   SUBPSRtoR( XMM0, XMM1 );

   if ( _X_Y_Z_W != 0xf )
   {
      MOVAPSMtoR( XMM2, VU_ACCx_ADDR );
      ANDPSMtoR( XMM0, (u32)&SSEmovMask[ _X_Y_Z_W ][ 0 ] );
      ANDPSMtoR( XMM2, (u32)&SSEmovMask[ 15 - _X_Y_Z_W ][ 0 ] );
      ORPSRtoR( XMM0, XMM2 );
   }

   MOVAPSRtoM( VU_ACCx_ADDR, XMM0 );
   recVUMI_STAT_REG_UPDATE( VU, VU_ACCx_ADDR );

   CPU_SSE_END;
}

void recVUMI_SUBAi( VURegs *VU ) { recVUMI_SUBA_iqxyzw( VU, VU_REGI_ADDR ); }
void recVUMI_SUBAq( VURegs *VU ) { recVUMI_SUBA_iqxyzw( VU, VU_REGQ_ADDR ); }
void recVUMI_SUBAx( VURegs *VU ) { recVUMI_SUBA_iqxyzw( VU, VU_VFx_ADDR( _Ft_ ) ); }
void recVUMI_SUBAy( VURegs *VU ) { recVUMI_SUBA_iqxyzw( VU, VU_VFy_ADDR( _Ft_ ) ); }
void recVUMI_SUBAz( VURegs *VU ) { recVUMI_SUBA_iqxyzw( VU, VU_VFz_ADDR( _Ft_ ) ); }
void recVUMI_SUBAw( VURegs *VU ) { recVUMI_SUBA_iqxyzw( VU, VU_VFw_ADDR( _Ft_ ) ); }

void recVUMI_MUL(VURegs *VU) {
	int t0reg;
	int fsreg;
	int ftreg;
	int fdreg;

	_addNeededVFtoXMMreg(VU, _Fs_);
	_addNeededVFtoXMMreg(VU, _Ft_);
	if (_Fd_) _addNeededVFtoXMMreg(VU, _Fd_);
	fsreg = _allocVFtoXMMreg(VU, -1, _Fs_, MODE_READ);
	ftreg = _allocVFtoXMMreg(VU, -1, _Ft_, MODE_READ);

	if (_X_Y_Z_W != 0xf) {
		if (_Fd_) fdreg = _allocVFtoXMMreg(VU, -1, _Fd_, MODE_READ | MODE_WRITE);
		else fdreg = _allocTempXMMreg(VU, -1);
		t0reg = _allocTempXMMreg(VU, -1);

		MOVAPSRtoR(t0reg, fsreg);
		MULPSRtoR(t0reg, ftreg);

		ANDPSMtoR(fdreg, (u32)&SSEmovMask[ 15 - _X_Y_Z_W ][ 0 ] );
		ANDPSMtoR(t0reg, (u32)&SSEmovMask[ _X_Y_Z_W ][ 0 ] );
		ORPSRtoR(fdreg, t0reg);

		_freeXMMreg(VU, t0reg);
	} else {
		if (_Fd_) fdreg = _allocVFtoXMMreg(VU, -1, _Fd_, MODE_WRITE);
		else fdreg = _allocTempXMMreg(VU, -1);

		MOVAPSRtoR(fdreg, fsreg);
		MULPSRtoR(fdreg, ftreg);
	}

	if (_Fd_ == 0) _freeXMMreg(VU, fdreg);
	_clearNeededXMMregs(VU);

	recVUMI_STAT_REG_UPDATE( VU, VU_ACCx_ADDR );
}

void recVUMI_MUL_iq(VURegs *VU, u32 addr) {
	int t0reg;
	int fsreg;
	int fdreg;

	_addNeededVFtoXMMreg(VU, _Fs_);
	if (_Fd_) _addNeededVFtoXMMreg(VU, _Fd_);
	t0reg = _allocTempXMMreg(VU, -1);
	fsreg = _allocVFtoXMMreg(VU, -1, _Fs_, MODE_READ);

	MOVSSMtoR(t0reg, addr); 
	UNPCKLPSRtoR(t0reg, t0reg);
	MOVLHPSRtoR(t0reg, t0reg);

	if (_X_Y_Z_W != 0xf) {
		if (_Fd_) fdreg = _allocVFtoXMMreg(VU, -1, _Fd_, MODE_READ | MODE_WRITE);
		else fdreg = _allocTempXMMreg(VU, -1);

		MULPSRtoR(t0reg, fsreg);

		ANDPSMtoR(fdreg, (u32)&SSEmovMask[ 15 - _X_Y_Z_W ][ 0 ] );
		ANDPSMtoR(t0reg, (u32)&SSEmovMask[ _X_Y_Z_W ][ 0 ] );
		ORPSRtoR(fdreg, t0reg);
	} else {
		if (_Fd_) fdreg = _allocVFtoXMMreg(VU, -1, _Fd_, MODE_WRITE);
		else fdreg = _allocTempXMMreg(VU, -1);

		MOVAPSRtoR(fdreg, fsreg);
		MULPSRtoR(fdreg, t0reg);
	}

	_freeXMMreg(VU, t0reg);
	if (_Fd_ == 0) _freeXMMreg(VU, fdreg);
	_clearNeededXMMregs(VU);

	recVUMI_STAT_REG_UPDATE( VU, VU_ACCx_ADDR );
}

void recVUMI_MUL_xyzw(VURegs *VU, int xyzw) {
	int t0reg;
	int fsreg;
	int ftreg;
	int fdreg;

	_addNeededVFtoXMMreg(VU, _Fs_);
	_addNeededVFtoXMMreg(VU, _Ft_);
	if (_Fd_) _addNeededVFtoXMMreg(VU, _Fd_);
	t0reg = _allocTempXMMreg(VU, -1);
	fsreg = _allocVFtoXMMreg(VU, -1, _Fs_, MODE_READ);
	ftreg = _allocVFtoXMMreg(VU, -1, _Ft_, MODE_READ);

	_unpackVF_xyzw(t0reg, ftreg, xyzw);

	if (_X_Y_Z_W != 0xf) {
		if (_Fd_) fdreg = _allocVFtoXMMreg(VU, -1, _Fd_, MODE_READ | MODE_WRITE);
		else fdreg = _allocTempXMMreg(VU, -1);

		MULPSRtoR(t0reg, fsreg);

		ANDPSMtoR(fdreg, (u32)&SSEmovMask[ 15 - _X_Y_Z_W ][ 0 ] );
		ANDPSMtoR(t0reg, (u32)&SSEmovMask[ _X_Y_Z_W ][ 0 ] );
		ORPSRtoR(fdreg, t0reg);
	} else {
		if (_Fd_) fdreg = _allocVFtoXMMreg(VU, -1, _Fd_, MODE_WRITE);
		else fdreg = _allocTempXMMreg(VU, -1);

		MOVAPSRtoR(fdreg, fsreg);
		MULPSRtoR(fdreg, t0reg);
	}

	_freeXMMreg(VU, t0reg);
	if (_Fd_ == 0) _freeXMMreg(VU, fdreg);
	_clearNeededXMMregs(VU);

	recVUMI_STAT_REG_UPDATE( VU, VU_ACCx_ADDR );
}

void recVUMI_MULi(VURegs *VU) { recVUMI_MUL_iq(VU, VU_REGI_ADDR); }
void recVUMI_MULq(VURegs *VU) { recVUMI_MUL_iq(VU, VU_REGQ_ADDR); }
void recVUMI_MULx(VURegs *VU) { recVUMI_MUL_xyzw(VU, 0); }
void recVUMI_MULy(VURegs *VU) { recVUMI_MUL_xyzw(VU, 1); }
void recVUMI_MULz(VURegs *VU) { recVUMI_MUL_xyzw(VU, 2); }
void recVUMI_MULw(VURegs *VU) { recVUMI_MUL_xyzw(VU, 3); }

void recVUMI_MULA( VURegs *VU ) {
	int t0reg;
	int accreg;
	int fsreg;
	int ftreg;

	_addNeededVFtoXMMreg(VU, _Fs_);
	_addNeededVFtoXMMreg(VU, _Ft_);
	_addNeededACCtoXMMreg(VU);
	fsreg = _allocVFtoXMMreg(VU, -1, _Fs_, MODE_READ);
	ftreg = _allocVFtoXMMreg(VU, -1, _Ft_, MODE_READ);

	if (_X_Y_Z_W != 0xf) {
		t0reg = _allocTempXMMreg(VU, -1);
		accreg = _allocACCtoXMMreg(VU, -1, MODE_READ | MODE_WRITE);

		MOVAPSRtoR(t0reg, fsreg);
		MULPSRtoR(t0reg, ftreg);

		ANDPSMtoR(accreg, (u32)&SSEmovMask[ 15 - _X_Y_Z_W ][ 0 ] );
		ANDPSMtoR(t0reg, (u32)&SSEmovMask[ _X_Y_Z_W ][ 0 ] );
		ORPSRtoR(accreg, t0reg);

		_freeXMMreg(VU, t0reg);
	} else {
		accreg = _allocACCtoXMMreg(VU, -1, MODE_WRITE);
		MOVAPSRtoR(accreg, fsreg);
		MULPSRtoR(accreg, ftreg);
	}

	_clearNeededXMMregs(VU);
	recVUMI_STAT_REG_UPDATE( VU, VU_ACCx_ADDR );
}

void recVUMI_MULA_iq(VURegs *VU, u32 addr) {
	int t0reg;
	int accreg;
	int fsreg;

	_addNeededVFtoXMMreg(VU, _Fs_);
	_addNeededACCtoXMMreg(VU);
	fsreg = _allocVFtoXMMreg(VU, -1, _Fs_, MODE_READ);

	if (_X_Y_Z_W != 0xf) {
		t0reg = _allocTempXMMreg(VU, -1);
		accreg = _allocACCtoXMMreg(VU, -1, MODE_READ | MODE_WRITE);

		MOVSSMtoR(t0reg, addr); 
		UNPCKLPSRtoR(t0reg, t0reg);
		MOVLHPSRtoR(t0reg, t0reg);

		MULPSRtoR(t0reg, fsreg);

		ANDPSMtoR(accreg, (u32)&SSEmovMask[15 - _X_Y_Z_W][0]);
		ANDPSMtoR(t0reg, (u32)&SSEmovMask[_X_Y_Z_W][0]);
		ORPSRtoR(accreg, t0reg);

		_freeXMMreg(VU, t0reg);
	} else {
		accreg = _allocACCtoXMMreg(VU, -1, MODE_WRITE);

		MOVSSMtoR(accreg, addr); 
		UNPCKLPSRtoR(accreg, accreg);
		MOVLHPSRtoR(accreg, accreg);

		MULPSRtoR(accreg, fsreg);
	}

	_clearNeededXMMregs(VU);
	recVUMI_STAT_REG_UPDATE(VU, VU_ACCx_ADDR);
}

void recVUMI_MULAi(VURegs *VU) { recVUMI_MULA_iq(VU, VU_REGI_ADDR); }
void recVUMI_MULAq(VURegs *VU) { recVUMI_MULA_iq(VU, VU_REGQ_ADDR); }

void recVUMI_MULA_xyzw(VURegs *VU, int xyzw) {
	int t0reg;
	int accreg;
	int fsreg;
	int ftreg;

	_addNeededVFtoXMMreg(VU, _Fs_);
	_addNeededVFtoXMMreg(VU, _Ft_);
	_addNeededACCtoXMMreg(VU);
	t0reg = _allocTempXMMreg(VU, -1);
	fsreg = _allocVFtoXMMreg(VU, -1, _Fs_, MODE_READ);
	ftreg = _allocVFtoXMMreg(VU, -1, _Ft_, MODE_READ);

	_unpackVF_xyzw(t0reg, ftreg, xyzw);
	MULPSRtoR(t0reg, fsreg);

	if (_X_Y_Z_W != 0xf) {
		accreg = _allocACCtoXMMreg(VU, -1, MODE_READ | MODE_WRITE);
		ANDPSMtoR(accreg, (u32)&SSEmovMask[15 - _X_Y_Z_W][0]);
		ANDPSMtoR(t0reg, (u32)&SSEmovMask[_X_Y_Z_W][0]);
		ORPSRtoR(accreg, t0reg);
	} else {
		accreg = _allocACCtoXMMreg(VU, -1, MODE_WRITE);
		MOVAPSRtoR(accreg, t0reg);
	}

	_clearNeededXMMregs(VU);
	recVUMI_STAT_REG_UPDATE(VU, VU_ACCx_ADDR);
}

void recVUMI_MULAx(VURegs *VU) { recVUMI_MULA_xyzw(VU, 0); }
void recVUMI_MULAy(VURegs *VU) { recVUMI_MULA_xyzw(VU, 1); }
void recVUMI_MULAz(VURegs *VU) { recVUMI_MULA_xyzw(VU, 2); }
void recVUMI_MULAw(VURegs *VU) { recVUMI_MULA_xyzw(VU, 3); }

void recVUMI_MADD(VURegs *VU) {
	int t0reg;
	int accreg;
	int fsreg;
	int ftreg;
	int fdreg;

	_addNeededVFtoXMMreg(VU, _Fs_);
	_addNeededVFtoXMMreg(VU, _Ft_);
	if (_Fd_) _addNeededVFtoXMMreg(VU, _Fd_);
	_addNeededACCtoXMMreg(VU);
	fsreg = _allocVFtoXMMreg(VU, -1, _Fs_, MODE_READ);
	ftreg = _allocVFtoXMMreg(VU, -1, _Ft_, MODE_READ);
	accreg = _allocACCtoXMMreg(VU, -1, MODE_READ);

	if (_X_Y_Z_W != 0xf) {
		if (_Fd_) fdreg = _allocVFtoXMMreg(VU, -1, _Fd_, MODE_READ | MODE_WRITE);
		else fdreg = _allocTempXMMreg(VU, -1);
		t0reg = _allocTempXMMreg(VU, -1);

		MOVAPSRtoR(t0reg, fsreg);
		MULPSRtoR(t0reg, ftreg);
		ADDPSRtoR(t0reg, accreg);

		ANDPSMtoR(fdreg, (u32)&SSEmovMask[ 15 - _X_Y_Z_W ][ 0 ] );
		ANDPSMtoR(t0reg, (u32)&SSEmovMask[ _X_Y_Z_W ][ 0 ] );
		ORPSRtoR(fdreg, t0reg);

		_freeXMMreg(VU, t0reg);
	} else {
		if (_Fd_) fdreg = _allocVFtoXMMreg(VU, -1, _Fd_, MODE_WRITE);
		else fdreg = _allocTempXMMreg(VU, -1);

		MOVAPSRtoR(fdreg, fsreg);
		MULPSRtoR(fdreg, ftreg);
		ADDPSRtoR(fdreg, accreg);
	}

	if (_Fd_ == 0) _freeXMMreg(VU, fdreg);
	_clearNeededXMMregs(VU);

	recVUMI_STAT_REG_UPDATE( VU, VU_ACCx_ADDR );
}

void recVUMI_MADD_iq(VURegs *VU, u32 addr) {
	int t0reg;
	int accreg;
	int fsreg;
	int fdreg;

	_addNeededVFtoXMMreg(VU, _Fs_);
	if (_Fd_) _addNeededVFtoXMMreg(VU, _Fd_);
	_addNeededACCtoXMMreg(VU);
	t0reg = _allocTempXMMreg(VU, -1);
	fsreg = _allocVFtoXMMreg(VU, -1, _Fs_, MODE_READ);
	accreg = _allocACCtoXMMreg(VU, -1, MODE_READ);

	MOVSSMtoR(t0reg, addr); 
	UNPCKLPSRtoR(t0reg, t0reg);
	MOVLHPSRtoR(t0reg, t0reg);

	if (_X_Y_Z_W != 0xf) {
		if (_Fd_) fdreg = _allocVFtoXMMreg(VU, -1, _Fd_, MODE_READ | MODE_WRITE);
		else fdreg = _allocTempXMMreg(VU, -1);

		MULPSRtoR(t0reg, fsreg);
		ADDPSRtoR(t0reg, accreg);

		ANDPSMtoR(fdreg, (u32)&SSEmovMask[ 15 - _X_Y_Z_W ][ 0 ] );
		ANDPSMtoR(t0reg, (u32)&SSEmovMask[ _X_Y_Z_W ][ 0 ] );
		ORPSRtoR(fdreg, t0reg);
	} else {
		if (_Fd_) fdreg = _allocVFtoXMMreg(VU, -1, _Fd_, MODE_WRITE);
		else fdreg = _allocTempXMMreg(VU, -1);

		MOVAPSRtoR(fdreg, fsreg);
		MULPSRtoR(fdreg, t0reg);
		ADDPSRtoR(fdreg, accreg);
	}

	_freeXMMreg(VU, t0reg);
	if (_Fd_ == 0) _freeXMMreg(VU, fdreg);
	_clearNeededXMMregs(VU);

	recVUMI_STAT_REG_UPDATE( VU, VU_ACCx_ADDR );
}

void recVUMI_MADD_xyzw(VURegs *VU, int xyzw) {
	int t0reg;
	int accreg;
	int fsreg;
	int ftreg;
	int fdreg;

	_addNeededVFtoXMMreg(VU, _Fs_);
	_addNeededVFtoXMMreg(VU, _Ft_);
	if (_Fd_) _addNeededVFtoXMMreg(VU, _Fd_);
	_addNeededACCtoXMMreg(VU);
	t0reg = _allocTempXMMreg(VU, -1);
	fsreg = _allocVFtoXMMreg(VU, -1, _Fs_, MODE_READ);
	ftreg = _allocVFtoXMMreg(VU, -1, _Ft_, MODE_READ);
	accreg = _allocACCtoXMMreg(VU, -1, MODE_READ);

	_unpackVF_xyzw(t0reg, ftreg, xyzw);

	if (_X_Y_Z_W != 0xf) {
		if (_Fd_) fdreg = _allocVFtoXMMreg(VU, -1, _Fd_, MODE_READ | MODE_WRITE);
		else fdreg = _allocTempXMMreg(VU, -1);

		MULPSRtoR(t0reg, fsreg);
		ADDPSRtoR(t0reg, accreg);

		ANDPSMtoR(fdreg, (u32)&SSEmovMask[ 15 - _X_Y_Z_W ][ 0 ] );
		ANDPSMtoR(t0reg, (u32)&SSEmovMask[ _X_Y_Z_W ][ 0 ] );
		ORPSRtoR(fdreg, t0reg);
	} else {
		if (_Fd_) fdreg = _allocVFtoXMMreg(VU, -1, _Fd_, MODE_WRITE);
		else fdreg = _allocTempXMMreg(VU, -1);

		MOVAPSRtoR(fdreg, fsreg);
		MULPSRtoR(fdreg, t0reg);
		ADDPSRtoR(fdreg, accreg);
	}

	_freeXMMreg(VU, t0reg);
	if (_Fd_ == 0) _freeXMMreg(VU, fdreg);
	_clearNeededXMMregs(VU);

	recVUMI_STAT_REG_UPDATE( VU, VU_ACCx_ADDR );
}

void recVUMI_MADDi(VURegs *VU) { recVUMI_MADD_iq(VU, VU_REGI_ADDR); }
void recVUMI_MADDq(VURegs *VU) { recVUMI_MADD_iq(VU, VU_REGQ_ADDR); }
void recVUMI_MADDx(VURegs *VU) { recVUMI_MADD_xyzw(VU, 0); }
void recVUMI_MADDy(VURegs *VU) { recVUMI_MADD_xyzw(VU, 1); }
void recVUMI_MADDz(VURegs *VU) { recVUMI_MADD_xyzw(VU, 2); }
void recVUMI_MADDw(VURegs *VU) { recVUMI_MADD_xyzw(VU, 3); }

void recVUMI_MADDA( VURegs *VU ) {
	int t0reg;
	int accreg;
	int fsreg;
	int ftreg;

	_addNeededVFtoXMMreg(VU, _Fs_);
	_addNeededVFtoXMMreg(VU, _Ft_);
	_addNeededACCtoXMMreg(VU);
	fsreg = _allocVFtoXMMreg(VU, -1, _Fs_, MODE_READ);
	ftreg = _allocVFtoXMMreg(VU, -1, _Ft_, MODE_READ);
	accreg = _allocACCtoXMMreg(VU, -1, MODE_READ | MODE_WRITE);
	t0reg = _allocTempXMMreg(VU, -1);

	MOVAPSRtoR(t0reg, fsreg);
	MULPSRtoR(t0reg, ftreg);

	if (_X_Y_Z_W != 0xf) {
		ADDPSRtoR(t0reg, accreg);

		ANDPSMtoR(accreg, (u32)&SSEmovMask[ 15 - _X_Y_Z_W ][ 0 ] );
		ANDPSMtoR(t0reg, (u32)&SSEmovMask[ _X_Y_Z_W ][ 0 ] );
		ORPSRtoR(accreg, t0reg);
	} else {
		ADDPSRtoR(accreg, t0reg);
	}

	_freeXMMreg(VU, t0reg);
	_clearNeededXMMregs(VU);
	recVUMI_STAT_REG_UPDATE( VU, VU_ACCx_ADDR );
}

void recVUMI_MADDA_iq( VURegs *VU, u32 addr) {
	int t0reg;
	int accreg;
	int fsreg;

	_addNeededVFtoXMMreg(VU, _Fs_);
	_addNeededACCtoXMMreg(VU);
	fsreg = _allocVFtoXMMreg(VU, -1, _Fs_, MODE_READ);
	accreg = _allocACCtoXMMreg(VU, -1, MODE_READ | MODE_WRITE);
	t0reg = _allocTempXMMreg(VU, -1);

	MOVSSMtoR(t0reg, addr); 
	UNPCKLPSRtoR(t0reg, t0reg);
	MOVLHPSRtoR(t0reg, t0reg);

	MULPSRtoR(t0reg, fsreg);

	if (_X_Y_Z_W != 0xf) {
		ADDPSRtoR(t0reg, accreg);

		ANDPSMtoR(accreg, (u32)&SSEmovMask[15 - _X_Y_Z_W][0]);
		ANDPSMtoR(t0reg, (u32)&SSEmovMask[_X_Y_Z_W][0]);
		ORPSRtoR(accreg, t0reg);

		_freeXMMreg(VU, t0reg);
	} else {
		ADDPSRtoR(accreg, t0reg);
	}

	_clearNeededXMMregs(VU);

	recVUMI_STAT_REG_UPDATE(VU, VU_ACCx_ADDR);
}

void recVUMI_MADDAi( VURegs *VU ) { recVUMI_MADDA_iq( VU, VU_REGI_ADDR ); }
void recVUMI_MADDAq( VURegs *VU ) { recVUMI_MADDA_iq( VU, VU_REGQ_ADDR ); }

void recVUMI_MADDA_xyzw(VURegs *VU, int xyzw) {
	int t0reg;
	int accreg;
	int fsreg;
	int ftreg;

	_addNeededVFtoXMMreg(VU, _Fs_);
	_addNeededVFtoXMMreg(VU, _Ft_);
	_addNeededACCtoXMMreg(VU);
	fsreg = _allocVFtoXMMreg(VU, -1, _Fs_, MODE_READ);
	ftreg = _allocVFtoXMMreg(VU, -1, _Ft_, MODE_READ);
	accreg = _allocACCtoXMMreg(VU, -1, MODE_READ | MODE_WRITE);
	t0reg = _allocTempXMMreg(VU, -1);
	_unpackVF_xyzw(t0reg, ftreg, xyzw);

	MULPSRtoR(t0reg, fsreg);

	if (_X_Y_Z_W != 0xf) {
		ADDPSRtoR(t0reg, accreg);

		ANDPSMtoR(accreg, (u32)&SSEmovMask[15 - _X_Y_Z_W][0]);
		ANDPSMtoR(t0reg, (u32)&SSEmovMask[_X_Y_Z_W][0]);
		ORPSRtoR(accreg, t0reg);

		_freeXMMreg(VU, t0reg);
	} else {
		ADDPSRtoR(accreg, t0reg);
	}

	_clearNeededXMMregs(VU);

	recVUMI_STAT_REG_UPDATE(VU, VU_ACCx_ADDR);
}

void recVUMI_MADDAx( VURegs *VU ) { recVUMI_MADDA_xyzw(VU, 0); }
void recVUMI_MADDAy( VURegs *VU ) { recVUMI_MADDA_xyzw(VU, 1); }
void recVUMI_MADDAz( VURegs *VU ) { recVUMI_MADDA_xyzw(VU, 2); }
void recVUMI_MADDAw( VURegs *VU ) { recVUMI_MADDA_xyzw(VU, 3); }

void recVUMI_MSUB( VURegs *VU )
{
   if ( _Fd_ == 0 ) return;
}

void recVUMI_MSUBi( VURegs *VU )
{
   if ( _Fd_ == 0 ) return;
}

void recVUMI_MSUBq( VURegs *VU )
{
   if ( _Fd_ == 0 ) return;
}

void recVUMI_MSUBx( VURegs *VU )
{
   if ( _Fd_ == 0 ) return;
}

void recVUMI_MSUBy( VURegs *VU )
{
   if ( _Fd_ == 0 ) return;
}

void recVUMI_MSUBz( VURegs *VU )
{
   if ( _Fd_ == 0 ) return;
}

void recVUMI_MSUBw( VURegs *VU )
{
   if ( _Fd_ == 0 ) return;
}

void recVUMI_MSUBA( VURegs *VU )
{
   if ( _Fd_ == 0 ) return;
}

void recVUMI_MSUBAi( VURegs *VU )
{
   if ( _Fd_ == 0 ) return;
}

void recVUMI_MSUBAq( VURegs *VU )
{
   if ( _Fd_ == 0 ) return;
}

void recVUMI_MSUBAx( VURegs *VU )
{
   if ( _Fd_ == 0 ) return;
}

void recVUMI_MSUBAy( VURegs *VU )
{
   if ( _Fd_ == 0 ) return;
}

void recVUMI_MSUBAz( VURegs *VU )
{
   if ( _Fd_ == 0 ) return;
}

void recVUMI_MSUBAw( VURegs *VU )
{
   if ( _Fd_ == 0 ) return;
}

void recVUMI_MAX(VURegs *VU) {
	int t0reg;
	int fsreg;
	int ftreg;
	int fdreg;

	if ( _Fd_ == 0 ) return;

	_addNeededVFtoXMMreg(VU, _Fs_);
	_addNeededVFtoXMMreg(VU, _Ft_);
	_addNeededVFtoXMMreg(VU, _Fd_);
	fsreg = _allocVFtoXMMreg(VU, -1, _Fs_, MODE_READ);
	ftreg = _allocVFtoXMMreg(VU, -1, _Ft_, MODE_READ);

	if (_X_Y_Z_W != 0xf) {
		fdreg = _allocVFtoXMMreg(VU, -1, _Fd_, MODE_READ | MODE_WRITE);
		t0reg = _allocTempXMMreg(VU, -1);

		MOVAPSRtoR(t0reg, fsreg);
		MAXPSRtoR(t0reg, ftreg);

		ANDPSMtoR(fdreg, (u32)&SSEmovMask[ 15 - _X_Y_Z_W ][ 0 ] );
		ANDPSMtoR(t0reg, (u32)&SSEmovMask[ _X_Y_Z_W ][ 0 ] );
		ORPSRtoR(fdreg, t0reg);

		_freeXMMreg(VU, t0reg);
	} else {
		fdreg = _allocVFtoXMMreg(VU, -1, _Fd_, MODE_WRITE);

		MOVAPSRtoR(fdreg, fsreg);
		MAXPSRtoR(fdreg, ftreg);
	}

	_clearNeededXMMregs(VU);

	recVUMI_STAT_REG_UPDATE( VU, VU_ACCx_ADDR );
}

void recVUMI_MAX_iq(VURegs *VU, u32 addr) {
	int t0reg;
	int fsreg;
	int fdreg;

	if ( _Fd_ == 0 ) return;

	_addNeededVFtoXMMreg(VU, _Fs_);
	_addNeededVFtoXMMreg(VU, _Fd_);
	t0reg = _allocTempXMMreg(VU, -1);
	fsreg = _allocVFtoXMMreg(VU, -1, _Fs_, MODE_READ);

	MOVSSMtoR(t0reg, addr); 
	UNPCKLPSRtoR(t0reg, t0reg);
	MOVLHPSRtoR(t0reg, t0reg);

	if (_X_Y_Z_W != 0xf) {
		fdreg = _allocVFtoXMMreg(VU, -1, _Fd_, MODE_READ | MODE_WRITE);

		MAXPSRtoR(t0reg, fsreg);

		ANDPSMtoR(fdreg, (u32)&SSEmovMask[ 15 - _X_Y_Z_W ][ 0 ] );
		ANDPSMtoR(t0reg, (u32)&SSEmovMask[ _X_Y_Z_W ][ 0 ] );
		ORPSRtoR(fdreg, t0reg);
	} else {
		fdreg = _allocVFtoXMMreg(VU, -1, _Fd_, MODE_WRITE);

		MOVAPSRtoR(fdreg, fsreg);
		MAXPSRtoR(fdreg, t0reg);
	}

	_freeXMMreg(VU, t0reg);
	_clearNeededXMMregs(VU);

	recVUMI_STAT_REG_UPDATE( VU, VU_ACCx_ADDR );
}

void recVUMI_MAX_xyzw(VURegs *VU, int xyzw) {
	int t0reg;
	int fsreg;
	int ftreg;
	int fdreg;

	if ( _Fd_ == 0 ) return;

	_addNeededVFtoXMMreg(VU, _Fs_);
	_addNeededVFtoXMMreg(VU, _Ft_);
	_addNeededVFtoXMMreg(VU, _Fd_);
	t0reg = _allocTempXMMreg(VU, -1);
	fsreg = _allocVFtoXMMreg(VU, -1, _Fs_, MODE_READ);
	ftreg = _allocVFtoXMMreg(VU, -1, _Ft_, MODE_READ);

	_unpackVF_xyzw(t0reg, ftreg, xyzw);

	if (_X_Y_Z_W != 0xf) {
		fdreg = _allocVFtoXMMreg(VU, -1, _Fd_, MODE_READ | MODE_WRITE);

		MAXPSRtoR(t0reg, fsreg);

		ANDPSMtoR(fdreg, (u32)&SSEmovMask[ 15 - _X_Y_Z_W ][ 0 ] );
		ANDPSMtoR(t0reg, (u32)&SSEmovMask[ _X_Y_Z_W ][ 0 ] );
		ORPSRtoR(fdreg, t0reg);
	} else {
		fdreg = _allocVFtoXMMreg(VU, -1, _Fd_, MODE_WRITE);

		MOVAPSRtoR(fdreg, fsreg);
		MAXPSRtoR(fdreg, t0reg);
	}

	_freeXMMreg(VU, t0reg);
	_clearNeededXMMregs(VU);

	recVUMI_STAT_REG_UPDATE( VU, VU_ACCx_ADDR );
}

void recVUMI_MAXi(VURegs *VU) { recVUMI_MAX_iq(VU, VU_REGI_ADDR); }
void recVUMI_MAXx(VURegs *VU) { recVUMI_MAX_xyzw(VU, 0); }
void recVUMI_MAXy(VURegs *VU) { recVUMI_MAX_xyzw(VU, 1); }
void recVUMI_MAXz(VURegs *VU) { recVUMI_MAX_xyzw(VU, 2); }
void recVUMI_MAXw(VURegs *VU) { recVUMI_MAX_xyzw(VU, 3); }

void recVUMI_MINI(VURegs *VU) {
	int t0reg;
	int fsreg;
	int ftreg;
	int fdreg;

	if ( _Fd_ == 0 ) return;

	_addNeededVFtoXMMreg(VU, _Fs_);
	_addNeededVFtoXMMreg(VU, _Ft_);
	_addNeededVFtoXMMreg(VU, _Fd_);
	fsreg = _allocVFtoXMMreg(VU, -1, _Fs_, MODE_READ);
	ftreg = _allocVFtoXMMreg(VU, -1, _Ft_, MODE_READ);

	if (_X_Y_Z_W != 0xf) {
		fdreg = _allocVFtoXMMreg(VU, -1, _Fd_, MODE_READ | MODE_WRITE);
		t0reg = _allocTempXMMreg(VU, -1);

		MOVAPSRtoR(t0reg, fsreg);
		MINPSRtoR(t0reg, ftreg);

		ANDPSMtoR(fdreg, (u32)&SSEmovMask[ 15 - _X_Y_Z_W ][ 0 ] );
		ANDPSMtoR(t0reg, (u32)&SSEmovMask[ _X_Y_Z_W ][ 0 ] );
		ORPSRtoR(fdreg, t0reg);

		_freeXMMreg(VU, t0reg);
	} else {
		fdreg = _allocVFtoXMMreg(VU, -1, _Fd_, MODE_WRITE);

		MOVAPSRtoR(fdreg, fsreg);
		MINPSRtoR(fdreg, ftreg);
	}

	_clearNeededXMMregs(VU);

	recVUMI_STAT_REG_UPDATE( VU, VU_ACCx_ADDR );
}

void recVUMI_MINI_iq(VURegs *VU, u32 addr) {
	int t0reg;
	int fsreg;
	int fdreg;

	if ( _Fd_ == 0 ) return;

	_addNeededVFtoXMMreg(VU, _Fs_);
	_addNeededVFtoXMMreg(VU, _Fd_);
	t0reg = _allocTempXMMreg(VU, -1);
	fsreg = _allocVFtoXMMreg(VU, -1, _Fs_, MODE_READ);

	MOVSSMtoR(t0reg, addr); 
	UNPCKLPSRtoR(t0reg, t0reg);
	MOVLHPSRtoR(t0reg, t0reg);

	if (_X_Y_Z_W != 0xf) {
		fdreg = _allocVFtoXMMreg(VU, -1, _Fd_, MODE_READ | MODE_WRITE);

		MINPSRtoR(t0reg, fsreg);

		ANDPSMtoR(fdreg, (u32)&SSEmovMask[ 15 - _X_Y_Z_W ][ 0 ] );
		ANDPSMtoR(t0reg, (u32)&SSEmovMask[ _X_Y_Z_W ][ 0 ] );
		ORPSRtoR(fdreg, t0reg);
	} else {
		fdreg = _allocVFtoXMMreg(VU, -1, _Fd_, MODE_WRITE);

		MOVAPSRtoR(fdreg, fsreg);
		MINPSRtoR(fdreg, t0reg);
	}

	_freeXMMreg(VU, t0reg);
	_clearNeededXMMregs(VU);

	recVUMI_STAT_REG_UPDATE( VU, VU_ACCx_ADDR );
}

void recVUMI_MINI_xyzw(VURegs *VU, int xyzw) {
	int t0reg;
	int fsreg;
	int ftreg;
	int fdreg;

	if ( _Fd_ == 0 ) return;

	_addNeededVFtoXMMreg(VU, _Fs_);
	_addNeededVFtoXMMreg(VU, _Ft_);
	_addNeededVFtoXMMreg(VU, _Fd_);
	t0reg = _allocTempXMMreg(VU, -1);
	fsreg = _allocVFtoXMMreg(VU, -1, _Fs_, MODE_READ);
	ftreg = _allocVFtoXMMreg(VU, -1, _Ft_, MODE_READ);

	_unpackVF_xyzw(t0reg, ftreg, xyzw);

	if (_X_Y_Z_W != 0xf) {
		fdreg = _allocVFtoXMMreg(VU, -1, _Fd_, MODE_READ | MODE_WRITE);

		MINPSRtoR(t0reg, fsreg);

		ANDPSMtoR(fdreg, (u32)&SSEmovMask[ 15 - _X_Y_Z_W ][ 0 ] );
		ANDPSMtoR(t0reg, (u32)&SSEmovMask[ _X_Y_Z_W ][ 0 ] );
		ORPSRtoR(fdreg, t0reg);
	} else {
		fdreg = _allocVFtoXMMreg(VU, -1, _Fd_, MODE_WRITE);

		MOVAPSRtoR(fdreg, fsreg);
		MINPSRtoR(fdreg, t0reg);
	}

	_freeXMMreg(VU, t0reg);
	_clearNeededXMMregs(VU);

	recVUMI_STAT_REG_UPDATE( VU, VU_ACCx_ADDR );
}

void recVUMI_MINIi(VURegs *VU) { recVUMI_MINI_iq(VU, VU_REGI_ADDR); }
void recVUMI_MINIx(VURegs *VU) { recVUMI_MINI_xyzw(VU, 0); }
void recVUMI_MINIy(VURegs *VU) { recVUMI_MINI_xyzw(VU, 1); }
void recVUMI_MINIz(VURegs *VU) { recVUMI_MINI_xyzw(VU, 2); }
void recVUMI_MINIw(VURegs *VU) { recVUMI_MINI_xyzw(VU, 3); }

void recVUMI_OPMULA( VURegs *VU )
{
   //VU->ACC.f.x = VU_MACx_UPDATE(VU, (double)VU->VF[_Fs_].f.y * VU->VF[_Ft_].f.z); 
   //VU->ACC.f.y = VU_MACy_UPDATE(VU, (double)VU->VF[_Fs_].f.z * VU->VF[_Ft_].f.x); 
   //VU->ACC.f.z = VU_MACz_UPDATE(VU, (double)VU->VF[_Fs_].f.x * VU->VF[_Ft_].f.y); 
	_freeXMMregs(VU);
   CPU_SSE_START;

   MOVAPSMtoR( XMM0, VU_VFx_ADDR( _Fs_ ) );
   MOVAPSMtoR( XMM1, VU_VFx_ADDR( _Ft_ ) );
   SHUFPSRtoR( XMM0, XMM0, 0xC9 );           // XMM0 = WXZY
   SHUFPSRtoR( XMM1, XMM1, 0xD2 );           // XMM1 = WYXZ
   MULPSRtoR( XMM0, XMM1 );
   MOVAPSRtoM( VU_ACCx_ADDR, XMM0 );

   recVUMI_STAT_REG_UPDATE( VU, VU_ACCx_ADDR );
   CPU_SSE_END;
}

void recVUMI_OPMSUB( VURegs *VU )
{
   if ( _Fd_ == 0 ) return;

	_freeXMMregs(VU);
   CPU_SSE_START;

   MOVAPSMtoR( XMM0, VU_ACCx_ADDR );
   MOVAPSMtoR( XMM1, VU_VFx_ADDR( _Fs_ ) );
   MOVAPSMtoR( XMM2, VU_VFx_ADDR( _Ft_ ) );
   SHUFPSRtoR( XMM1, XMM1, 0xC9 );
   SHUFPSRtoR( XMM2, XMM2, 0xD2 );
   MULPSRtoR( XMM1, XMM2 );
   SUBPSRtoR( XMM0, XMM1 );
   MOVAPSRtoM( VU_VFx_ADDR( _Fd_ ), XMM0 );

   recVUMI_STAT_REG_UPDATE( VU, VU_VFx_ADDR( _Fd_ ) );

   CPU_SSE_END;
}

void recVUMI_NOP( VURegs *VU ) 
{
}

void recVUMI_FTOI0(VURegs *VU) {
	if ( _Ft_ == 0 ) return; 

	_freeXMMregs(VU);
   if ( _X_Y_Z_W != 0xf )
   {
      CVTPS2DQMtoR( XMM0, VU_VFx_ADDR( _Fs_ ) );
      MOVAPSMtoR( XMM1, VU_VFx_ADDR( _Ft_ ) );
      ANDPSMtoR( XMM0, (u32)&SSEmovMask[ _X_Y_Z_W ][ 0 ] );
      ANDPSMtoR( XMM1, (u32)&SSEmovMask[ 15 - _X_Y_Z_W ][ 0 ] );
      ORPSRtoR( XMM0, XMM1 );
      MOVAPSRtoM( VU_VFx_ADDR( _Ft_ ), XMM0 );
   }
   else
   {
      CVTPS2DQMtoR( XMM0, VU_VFx_ADDR( _Fs_ ) );
      MOVAPSRtoM( VU_VFx_ADDR( _Ft_ ), XMM0 );
   }
}

void recVUMI_FTOI4( VURegs *VU )
{
	if ( _Ft_ == 0 ) return; 

	_freeXMMregs(VU);
   CPU_SSE_START;

   MOVAPSMtoR( XMM1, (u32)&recMult_float_to_int4[0] );

   if ( _X_Y_Z_W != 0xf )
   {
      MOVAPSMtoR( XMM0, VU_VFx_ADDR( _Ft_ ) );
      MULPSRtoR( XMM1, XMM0 );
      ANDNPSMtoR( XMM0, (u32)&SSEmovMask[ _X_Y_Z_W ][ 0 ] );
      ANDPSMtoR( XMM1, (u32)&SSEmovMask[ _X_Y_Z_W ][ 0 ] );
      ORPSRtoR( XMM0, XMM1 );
   }
   else
   {
      MOVAPSMtoR( XMM0, VU_VFx_ADDR( _Fs_ ) );
      MULPSRtoR( XMM0, XMM1 );
   }

   MOVAPSRtoM( VU_VFx_ADDR( _Ft_ ), XMM0 );

   CPU_SSE_END;
}

void recVUMI_FTOI12( VURegs *VU )
{
   if ( _Ft_ == 0 ) return;
}

void recVUMI_FTOI15( VURegs *VU )
{
   if ( _Ft_ == 0 ) return;
}

void recVUMI_ITOF0( VURegs *VU )
{
   if ( _Ft_ == 0 ) return;

	_freeXMMregs(VU);
   if ( _X_Y_Z_W != 0xf )
   {
      CVTDQ2PSMtoR( XMM0, VU_VFx_ADDR( _Fs_ ) );
      MOVAPSMtoR( XMM1, VU_VFx_ADDR( _Ft_ ) );
      ANDPSMtoR( XMM0, (u32)&SSEmovMask[ _X_Y_Z_W ][ 0 ] );
      ANDPSMtoR( XMM1, (u32)&SSEmovMask[ 15 - _X_Y_Z_W ][ 0 ] );
      ORPSRtoR( XMM0, XMM1 );
      MOVAPSRtoM( VU_VFx_ADDR( _Ft_ ), XMM0 );
   }
   else
   {
      CVTDQ2PSMtoR( XMM0, VU_VFx_ADDR( _Fs_ ) );
      MOVAPSRtoM( VU_VFx_ADDR( _Ft_ ), XMM0 );
   }

}

void recVUMI_ITOF4( VURegs *VU )
{
   if ( _Ft_ == 0 ) return;
}

void recVUMI_ITOF12( VURegs *VU )
{
   if ( _Ft_ == 0 ) return;
}

void recVUMI_ITOF15( VURegs *VU )
{
   if ( _Ft_ == 0 ) return;
}

static u32 const_clipAND[4] = 
   { 0x7fffffff, 0x7fffffff, 0x7fffffff, 0x7fffffff };
static u32 const_clipOR[4] = 
   { 0x80000000, 0x80000000, 0x80000000, 0x80000000 };

void recVUMI_CLIP(VURegs *VU) {
	int t0reg;
	int fsreg;
	int ftreg;

	_addNeededVFtoXMMreg(VU, _Fs_);
	_addNeededVFtoXMMreg(VU, _Ft_);
	t0reg = _allocTempXMMreg(VU, -1);
	fsreg = _allocVFtoXMMreg(VU, -1, _Fs_, MODE_READ);
	ftreg = _allocVFtoXMMreg(VU, -1, _Ft_, MODE_READ);

	MOV32MtoR(EAX, (uptr)&VU->clipflag);
	SHL32ItoR(EAX, 6);

	_unpackVF_xyzw(t0reg, ftreg, 3);
	ANDPSMtoR(t0reg, (u32)const_clipAND);
	CMPNLEPSRtoR(t0reg, fsreg);
//EXTRW();
	MOVD64XMMtoR(EDX, t0reg);
	OR32RtoR(EAX, EDX);
	MOVD64XMMtoR(EDX, t0reg);
	OR32RtoR(EAX, EDX);

	_unpackVF_xyzw(t0reg, ftreg, 3);
	ORPSMtoR(t0reg, (u32)const_clipOR);
	CMPLTPSRtoR(t0reg, fsreg);


	MOV32RtoM((uptr)&VU->clipflag, EAX);
	MOV32RtoM((uptr)&VU->VI[REG_CLIP_FLAG].UL, EAX);

	_freeXMMreg(VU, t0reg);
	_clearNeededXMMregs(VU);
}

/******************************/
/*   VU Lower instructions    */
/******************************/

void recVUMI_DIV(VURegs *VU) {
	_freeXMMregs(VU);

	SetFPUstate( );

	MOV32MtoR( EAX, (u32)&VU->VF[ _Ft_ ].F[ _Ftf_ ] );
	OR32RtoR( EAX, EAX );
	j8Ptr[ 0 ] = JNE8( 0 );
	MOV32MtoR( EAX, (u32)&VU->VF[ _Fs_ ].F[ _Fsf_ ] );
	OR32RtoR( EAX, EAX );
	j8Ptr[ 1 ] = JNE8( 0 );
	recVU_UPDATE_STAT_REG_DI2( VU );
	j8Ptr[ 2 ] = JMP8( 0 );
	x86SetJ8( j8Ptr[ 1 ] );
	recVU_UPDATE_STAT_REG_DI1( VU );
	j8Ptr[ 3 ] = JMP8( 0 );
	x86SetJ8( j8Ptr[ 0 ] );
	recVU_UPDATE_STAT_REG_DI0( VU );

	FLD32( (u32)&VU->VF[ _Fs_ ].F[ _Fsf_ ] );
	FDIV32( (u32)&VU->VF[ _Ft_ ].F[ _Ftf_ ] );
	FSTP32( (u32)&VU->fdiv.reg.F );
	_recvuFDIVAdd(VU, 6);

	x86SetJ8( j8Ptr[ 2 ] );
	x86SetJ8( j8Ptr[ 3 ] );
}

void recVUMI_SQRT( VURegs *VU ) {
	_freeXMMregs(VU);

	FLD32((u32)&VU->VF[ _Ft_ ].F[ _Ftf_ ]);
	FSQRT();
	FSTP32( (u32)&VU->fdiv.reg.F );
	_recvuFDIVAdd(VU, 6);
}

static u32 rsqrttmp;

void recVUMI_RSQRT(VURegs *VU) {
	SysPrintf("RSQRT\n");
	_freeXMMregs(VU);

	FLD32((u32)&VU->VF[ _Ft_ ].F[ _Ftf_ ]);
	FSQRT( );
	FSTP32( (u32)&rsqrttmp );

	MOV32MtoR( EAX, (u32)&rsqrttmp );
	OR32RtoR( EAX, EAX );
	j8Ptr[ 0 ] = JE8( 0 );				
	FLD32( (u32)&VU->VF[ _Fs_ ].F[ _Fsf_ ] );
	FDIV32( (u32)&rsqrttmp );
	FSTP32( (u32)&VU->fdiv.reg.F );
	x86SetJ8( j8Ptr[ 0 ] );

	_recvuFDIVAdd(VU, 12);
}

void _addISIMMtoIT(VURegs *VU, s16 imm) {
	if (_Ft_ == _Fs_) {
		if (imm != 0 ) {
			ADD16ItoM(VU_VI_ADDR(_Ft_ ), imm);
		}
	} else {
		if (_Fs_ == 0) {
			MOV16ItoM(VU_VI_ADDR(_Ft_), imm);
		} else {
			MOV16MtoR(EAX, VU_VI_ADDR( _Fs_ ) );
			if (imm != 0) {
				ADD16ItoR( EAX, imm );
			}
			MOV16RtoM(VU_VI_ADDR( _Ft_ ), EAX );
		}
	}
}

void recVUMI_IADDI(VURegs *VU) {
	s16 imm;

	if ( _Ft_ == 0 ) return;

	imm = ( VU->code >> 6 ) & 0x1f;
	imm = ( imm & 0x10 ? 0xfff0 : 0) | ( imm & 0xf );
	_addISIMMtoIT(VU, imm);
}

void recVUMI_IADDIU(VURegs *VU) {
	u32 imm;

	if ( _Ft_ == 0 ) return;

	imm = ( ( VU->code >> 10 ) & 0x7800 ) | ( VU->code & 0x7ff );
	_addISIMMtoIT(VU, imm);
}

void recVUMI_IADD( VURegs *VU )
{
   if ( _Fd_ == 0 ) return;

   if ( ( _Ft_ == 0 ) && ( _Fs_ == 0 ) )
   {
      MOV16ItoM( VU_VI_ADDR( _Fd_ ), 0 );
   } 
   else if ( _Fs_ == 0 )
   {
      MOV16MtoR( EAX, VU_VI_ADDR( _Ft_ ) );
      MOV16RtoM( VU_VI_ADDR( _Fd_ ), EAX );
   }
   else if ( _Ft_ == 0 )
   {
      MOV16MtoR( EAX, VU_VI_ADDR( _Fs_ ) );
      MOV16RtoM( VU_VI_ADDR( _Fd_ ), EAX );
   }
   else
   {
      MOV16MtoR( EAX, VU_VI_ADDR( _Fs_ ) );
      ADD16MtoR( EAX, VU_VI_ADDR( _Ft_ ) );
      MOV16RtoM( VU_VI_ADDR( _Fd_ ), EAX );
   }
}

void recVUMI_IAND( VURegs *VU )
{
   if ( _Fd_ == 0 ) return;

   if ( ( _Fs_ == 0 ) || ( _Ft_ == 0 ) )
   {
      XOR16RtoR( EAX, EAX );
      MOV16RtoM( VU_VI_ADDR( _Fd_ ), EAX );
   }
   else
   {
      if ( _Fs_ == _Ft_ )
      {
         MOV16MtoR( EAX, VU_VI_ADDR( _Fs_ ) );
         MOV16RtoM( VU_VI_ADDR( _Fd_ ), EAX );
      }
      else if ( _Fd_ == _Fs_ )
      {
         MOV16MtoR( EAX, VU_VI_ADDR( _Ft_ ) );
         AND16RtoM( VU_VI_ADDR( _Fd_ ), EAX );
      }
      else if ( _Fd_ == _Ft_ )
      {
         MOV16MtoR( EAX, VU_VI_ADDR( _Fs_ ) );
         AND16RtoM( VU_VI_ADDR( _Fd_ ), EAX );
      }
      else
      {
         MOV16MtoR( EAX, VU_VI_ADDR( _Fs_ ) );
         AND16MtoR( EAX, VU_VI_ADDR( _Ft_ ) );
         MOV16RtoM( VU_VI_ADDR( _Fd_ ), EAX );
      }
   }
}

void recVUMI_IOR( VURegs *VU )
{
   if ( _Fd_ == 0 ) return;

   if ( _Fs_ == 0 )
   {
      MOV16MtoR( EAX, VU_VI_ADDR( _Ft_ ) );
      MOV16RtoM( VU_VI_ADDR( _Fd_ ), EAX );
   }
   else
   {
      MOV16MtoR( EAX, VU_VI_ADDR( _Fs_ ) );
      if ( _Ft_ != 0 )
      {
         OR16MtoR( EAX, VU_VI_ADDR( _Ft_ ) );
      }
      MOV16RtoM( VU_VI_ADDR( _Fd_ ), EAX );
   }
}

void recVUMI_ISUB( VURegs *VU )
{
   if ( _Fd_ == 0 ) return;

   if ( _Fs_ == 0 )
   {
      MOV16MtoR( EAX, VU_VI_ADDR( _Ft_ ) );
      NEG16R( EAX );
      MOV16RtoM( VU_VI_ADDR( _Fd_ ), EAX );
      return;
   }
   else if ( _Ft_ == 0 )
   {
      MOV16MtoR( EAX, VU_VI_ADDR( _Fs_ ) );
      MOV16RtoM( VU_VI_ADDR( _Fd_ ), EAX );
   }
   else
   {
      MOV16MtoR( EAX, VU_VI_ADDR( _Fs_ ) );
      SUB16MtoR( EAX, VU_VI_ADDR( _Ft_ ) );
      MOV16RtoM( VU_VI_ADDR( _Fd_ ), EAX );
   }
}

void recVUMI_ISUBIU( VURegs *VU )
{
   if( _Ft_ == 0) return;
}

void recVUMI_MOVE( VURegs *VU )
{
   if ( _Ft_ == 0 ) return;

	_freeXMMregs(VU);
   CPU_SSE_START;
   
   MOVAPSMtoR( XMM0, VU_VFx_ADDR( _Fs_ ) );
   if ( _X_Y_Z_W != 0xf )
   {
      MOVAPSMtoR( XMM2, VU_VFx_ADDR( _Ft_ ) );
      ANDPSMtoR( XMM0, (u32)&SSEmovMask[ _X_Y_Z_W ][ 0 ] );
      ANDPSMtoR( XMM2, (u32)&SSEmovMask[ 15 - _X_Y_Z_W ][ 0 ] );
      ORPSRtoR( XMM0, XMM2 );
   }
   MOVAPSRtoM( VU_VFx_ADDR( _Ft_ ), XMM0 );
   
   CPU_SSE_END;
}

void recVUMI_MFIR( VURegs *VU ) {
	int t0reg;
	int ftreg;

	if ( _Ft_ == 0 ) return;

	_addNeededVFtoXMMreg(VU, _Ft_);
	t0reg = _allocTempXMMreg(VU, -1);

	MOVSX32M16toR(EAX, VU_VI_ADDR(_Fs_));
	MOVD32RtoXMM(t0reg, EAX);
	UNPCKLPSRtoR(t0reg, t0reg);
	MOVLHPSRtoR(t0reg, t0reg);

    if (_X_Y_Z_W != 0xf) {
		ftreg = _allocVFtoXMMreg(VU, -1, _Ft_, MODE_READ | MODE_WRITE);

		ANDPSMtoR(ftreg, (u32)&SSEmovMask[15 - _X_Y_Z_W][0]);
		ANDPSMtoR(t0reg, (u32)&SSEmovMask[_X_Y_Z_W][0]);

		ORPSRtoR(ftreg, t0reg);
    } else {
		ftreg = _allocVFtoXMMreg(VU, -1, _Ft_, MODE_WRITE);
		MOVAPSRtoR(ftreg, t0reg);
	}

	_freeXMMreg(VU, t0reg);
	_clearNeededXMMregs(VU);
}

void recVUMI_MTIR( VURegs *VU )
{
   if ( _Ft_ == 0 ) return;

	_freeXMMregs(VU);

   MOV32MtoR( EAX, (u32)&VU->VF[ _Fs_ ].F[ _Fsf_ ] );
   MOV16RtoM( VU_VI_ADDR( _Ft_ ), EAX );
} 

void recVUMI_MR32( VURegs *VU ) {
   if ( _Ft_ == 0 ) return;

	_freeXMMregs(VU);

   MOVAPSMtoR( XMM0, VU_VFx_ADDR( _Fs_ ) );

   if ( _X_Y_Z_W != 0xf )
   {
      PSHUFDRtoR(XMM0, XMM0, ((0) << 6) | ((3) << 4) | ((2) << 2) | (1));

      MOVAPSMtoR( XMM1, VU_VFx_ADDR( _Ft_ ) );
      ANDPSMtoR( XMM0, (u32)&SSEmovMask[ _X_Y_Z_W ][ 0 ] );
      ANDPSMtoR( XMM1, (u32)&SSEmovMask[ 15 - _X_Y_Z_W ][ 0 ] );
		ORPSRtoR(XMM0, XMM1);
   } else {
      PSHUFDMtoR(XMM0, VU_VFx_ADDR( _Ft_ ), ((0) << 6) | ((3) << 4) | ((2) << 2) | (1));
   }

   MOVAPSRtoM( VU_VFx_ADDR( _Ft_ ), XMM0 );
}

void _loadEAX(VURegs *VU) {
	int t0reg;
	int ftreg;

    if (_X_Y_Z_W != 0xf) {
		ftreg = _allocVFtoXMMreg(VU, -1, _Ft_, MODE_READ | MODE_WRITE);
		t0reg = _allocTempXMMreg(VU, -1);

	    MOVAPSRmtoR(t0reg, EAX);
		ANDPSMtoR(ftreg, (u32)&SSEmovMask[15 - _X_Y_Z_W][0]);
		ANDPSMtoR(t0reg, (u32)&SSEmovMask[_X_Y_Z_W][0]);

		ORPSRtoR(ftreg, t0reg);
		_freeXMMreg(VU, t0reg);
    } else {
		ftreg = _allocVFtoXMMreg(VU, -1, _Ft_, MODE_WRITE);
	    MOVAPSRmtoR(ftreg, EAX);
	}

	_clearNeededXMMregs(VU);
}

void recVUMI_LQ(VURegs *VU) {
	s16 imm;

	if ( _Ft_ == 0 ) return;

	imm = ( VU->code & 0x400) ? ( VU->code & 0x3ff) | 0xfc00 : ( VU->code & 0x3ff); 
	if ( _Fs_ == 0 ) {
		MOV32ItoR( EAX, (u32)VU->Mem + ((imm * 16) & 0x3fff) );
	} else {
		MOVZX32M16toR( EAX, VU_VI_ADDR( _Fs_ ) );
		SHL32ItoR( EAX, 4 );
		ADD32ItoR( EAX, imm * 16 );
		AND32ItoR( EAX, 0x3fff);
		ADD32ItoR( EAX, (u32)VU->Mem );
	}

	_loadEAX(VU);
}

void recVUMI_LQD( VURegs *VU ) {
	if ( _Fs_ != 0 ) {
		DEC16M( VU_VI_ADDR( _Fs_ ) );
	}

	if ( _Ft_ == 0 ) return;

	if ( _Fs_ == 0 ) {
		MOV32ItoR( EAX, (u32)VU->Mem );
	} else {
		MOVZX32M16toR( EAX, VU_VI_ADDR( _Fs_ ) );
		SHL32ItoR( EAX, 4 );
		AND32ItoR( EAX, 0x3fff);
		ADD32ItoR( EAX, (u32)VU->Mem );
	}

	_loadEAX(VU);
}

void recVUMI_LQI(VURegs *VU) {
	if ( _Ft_ == 0 ) {
		INC32M( VU_VI_ADDR( _Fs_ ) );
		return;
	}

    if (_Fs_ == 0) {
		MOV32ItoR( EAX, (u32)VU->Mem );
    } else {
		MOVZX32M16toR( EAX, VU_VI_ADDR( _Fs_ ) );
		SHL32ItoR( EAX, 4 );
		AND32ItoR( EAX, 0x3fff);
		ADD32ItoR( EAX, (u32)VU->Mem );
		INC32M( VU_VI_ADDR( _Fs_ ) );
    }

	_loadEAX(VU);
}

void _saveEAX(VURegs *VU) {
	int t0reg;
	int fsreg;

	fsreg = _allocVFtoXMMreg(VU, -1, _Fs_, MODE_READ);

    if (_X_Y_Z_W != 0xf) {
		t0reg = _allocTempXMMreg(VU, -1);

	    MOVAPSRmtoR(t0reg, EAX);
		ANDPSMtoR(fsreg, (u32)&SSEmovMask[_X_Y_Z_W][0]);
		ANDPSMtoR(t0reg, (u32)&SSEmovMask[15 - _X_Y_Z_W][0]);

		ORPSRtoR(t0reg, fsreg);
		MOVAPSRtoRm(EAX, t0reg);

		_freeXMMreg(VU, t0reg);
    } else {
		MOVAPSRtoRm(EAX, fsreg);
	}

	_clearNeededXMMregs(VU);
}

void recVUMI_SQ(VURegs *VU) {
	s16 imm;
 
	imm = ( VU->code & 0x400) ? ( VU->code & 0x3ff) | 0xfc00 : ( VU->code & 0x3ff); 
	if ( _Ft_ == 0 ) {
      MOV32ItoR( EAX, (u32)VU->Mem + imm * 16 );
	} else {
		MOVZX32M16toR( EAX, VU_VI_ADDR( _Ft_ ) );
		SHL32ItoR( EAX, 4 );
		ADD32ItoR( EAX, imm * 16 );
		AND32ItoR( EAX, 0x3fff);
		ADD32ItoR( EAX, (u32)VU->Mem );
	}

	_saveEAX(VU);
}

void recVUMI_SQD(VURegs *VU) {
	if (_Ft_ == 0) {
		MOV32ItoR( EAX, (u32)VU->Mem );
	} else {
		SUB32ItoM( VU_VI_ADDR( _Ft_ ), 1 );
		MOVZX32M16toR( EAX, VU_VI_ADDR( _Ft_ ) );
		SHL32ItoR( EAX, 4 );
		AND32ItoR( EAX, 0x3fff);
		ADD32ItoR( EAX, (u32)VU->Mem );
	}

	_saveEAX(VU);
}

void recVUMI_SQI(VURegs *VU) {
	if (_Ft_ == 0) {
		MOV32ItoR( EAX, (u32)VU->Mem );
	} else {
		MOVZX32M16toR( EAX, VU_VI_ADDR( _Ft_ ) );
		SHL32ItoR( EAX, 4 );
		AND32ItoR( EAX, 0x3fff);
		ADD32ItoR( EAX, (u32)VU->Mem );
		ADD32ItoM( VU_VI_ADDR( _Ft_ ), 1 );
	}

	_saveEAX(VU);
}

void recVUMI_ILW(VURegs *VU) {
	s16 imm;
 
	if ( _Ft_ == 0 ) return;

	imm = ( VU->code & 0x400) ? ( VU->code & 0x3ff) | 0xfc00 : ( VU->code & 0x3ff); 
	if ( _Ft_ == 0 ) {
		MOV32ItoR( EAX, (u32)VU->Mem + imm * 16 );
	} else {
		MOVZX32M16toR( EAX, VU_VI_ADDR( _Fs_ ) );
		SHL32ItoR( EAX, 4 );
		ADD32ItoR( EAX, imm * 16 );
		AND32ItoR( EAX, 0x3fff);
		ADD32ItoR( EAX, (u32)VU->Mem );
	}

	if (_X) XOR32RtoR(EDX, EDX);
	else if (_Y) MOV32ItoR(EDX, 1);
	else if (_Z) MOV32ItoR(EDX, 2);
	else if (_W) MOV32ItoR(EDX, 3);
	MOV32RmStoR( ECX, EAX, EDX, 2 );
	MOV16RtoM( (u32)&VU->VI[ _Ft_ ].US[ 0 ], ECX );
}

void recVUMI_ISW( VURegs *VU ) {
	s16 imm;
 
	if ( _Ft_ == 0 ) return;

	imm = ( VU->code & 0x400) ? ( VU->code & 0x3ff) | 0xfc00 : ( VU->code & 0x3ff); 
	if (_Ft_ == 0) {
		MOV32ItoR( EAX, (u32)VU->Mem + imm * 16 );
	} else {
		MOVZX32M16toR( EAX, VU_VI_ADDR( _Fs_ ) );
		SHL32ItoR( EAX, 4 );
		ADD32ItoR( EAX, imm * 16 );
		AND32ItoR( EAX, 0x3fff);
		ADD32ItoR( EAX, (u32)VU->Mem );
	}

	MOV32MtoR( ECX, (u32)&VU->VI[ _Ft_ ].US[ 0 ] );
	if (_X) XOR32RtoR(EDX, EDX);
	else if (_Y) MOV32ItoR(EDX, 1);
	else if (_Z) MOV32ItoR(EDX, 2);
	else if (_W) MOV32ItoR(EDX, 3);
	MOV16RtoRmS( ECX, EAX, EDX, 2 );
}

void recVUMI_ILWR( VURegs *VU ) {
	if ( _Ft_ == 0 ) return;

	if ( _Fs_ == 0 ) {
		MOV32ItoR( EAX, (u32)VU->Mem );
	} else {
		MOVZX32M16toR( EAX, VU_VI_ADDR( _Fs_ ) );
		SHL32ItoR( EAX, 4 );
		AND32ItoR( EAX, 0x3fff);
		ADD32ItoR( EAX, (u32)VU->Mem );
	}

	if (_X) XOR32RtoR(EDX, EDX);
	else if (_Y) MOV32ItoR(EDX, 1);
	else if (_Z) MOV32ItoR(EDX, 2);
	else if (_W) MOV32ItoR(EDX, 3);
	MOV32RmStoR( ECX, EAX, EDX, 2 );
	MOV16RtoM( (u32)&VU->VI[ _Ft_ ].US[ 0 ], ECX );
}

void recVUMI_ISWR( VURegs *VU ) {
	if ( _Ft_ == 0 ) return;

	if ( _Fs_ == 0 ) {
		MOV32ItoR( EAX, (u32)VU->Mem );
	} else {
		MOVZX32M16toR( EAX, VU_VI_ADDR( _Fs_ ) );
		SHL32ItoR( EAX, 4 );
		AND32ItoR( EAX, 0x3fff);
		ADD32ItoR( EAX, (u32)VU->Mem );
	}

	MOV32MtoR( ECX, (u32)&VU->VI[ _Ft_ ].US[ 0 ] );
	if (_X) XOR32RtoR(EDX, EDX);
	else if (_Y) MOV32ItoR(EDX, 1);
	else if (_Z) MOV32ItoR(EDX, 2);
	else if (_W) MOV32ItoR(EDX, 3);
	MOV16RtoRmS( ECX, EAX, EDX, 2 );
}

void recVUMI_RINIT(VURegs *VU) {
	_freeXMMregs(VU);
	MOV32MtoR( EAX, VU_VFx_ADDR( _Fs_ ) + 4 * _Fsf_ );
	AND32ItoR( EAX, 0x7fffff );
	OR32ItoR( EAX, 0x7f << 23 );
	MOV32RtoM( VU_REGR_ADDR, EAX );
}

void recVUMI_RGET(VURegs *VU) {
   if ( _Ft_ == 0 ) return;

	_freeXMMregs(VU);

   MOV32MtoR( EAX, VU_REGR_ADDR );
   if ( _X )   MOV32RtoM( VU_VFx_ADDR( _Ft_ ), EAX );
   if ( _Y )   MOV32RtoM( VU_VFy_ADDR( _Ft_ ), EAX );
   if ( _Z )   MOV32RtoM( VU_VFz_ADDR( _Ft_ ), EAX );
   if ( _W )   MOV32RtoM( VU_VFw_ADDR( _Ft_ ), EAX );
}

void recVUMI_RNEXT( VURegs *VU )
{
   if ( _Ft_ == 0) return;

	_freeXMMregs(VU);

   MOV32MtoR( EAX, VU_REGR_ADDR );
   MOV32ItoR( ECX, 0x15a4e35 );
   IMUL32R( ECX );
   INC32R( EAX );
   SHR32ItoR( EAX, 9 );
   MOV32RtoM( VU_REGR_ADDR, EAX );
   OR32ItoR( EAX, 0x7f << 23 );
   if ( _X )   MOV32RtoM( VU_VFx_ADDR( _Ft_ ), EAX );
   if ( _Y )   MOV32RtoM( VU_VFy_ADDR( _Ft_ ), EAX );
   if ( _Z )   MOV32RtoM( VU_VFz_ADDR( _Ft_ ), EAX );
   if ( _W )   MOV32RtoM( VU_VFw_ADDR( _Ft_ ), EAX );
}

void recVUMI_RXOR( VURegs *VU )
{
	_freeXMMregs(VU);

   MOV32MtoR( EAX, VU_VFx_ADDR( _Fs_ ) + 4 * _Fsf_ );
   XOR32MtoR( EAX, VU_REGR_ADDR );
   AND32ItoR( EAX, 0x7fffff );
   OR32ItoR ( EAX, 0x7f << 23 );
   MOV32RtoM( VU_REGR_ADDR, EAX );
} 

void recVUMI_WAITQ( VURegs *VU )
{
}

void recVUMI_FSAND( VURegs *VU )
{
   if ( _Ft_ == 0 ) return;
}

void recVUMI_FSEQ( VURegs *VU )
{
   if ( _Ft_ == 0 ) return;
}

void recVUMI_FSOR( VURegs *VU )
{
   if ( _Ft_ == 0 ) return;
}

void recVUMI_FSSET( VURegs *VU )
{
}

void recVUMI_FMAND( VURegs *VU )
{
   if ( _Ft_ == 0 ) return;

   recUpdateFlags( );

   MOV32MtoR( EAX, (u32)&VU->VI[ _Fs_ ].US[ 0 ] );
   AND32MtoR( EAX, (u32)&VU->VI[ REG_MAC_FLAG ].US[ 0 ] );
   AND32ItoR( EAX, 0xFFFF );
   MOV32RtoM( (u32)&VU->VI[ _Ft_ ].US[ 0 ], EAX );
}

void recVUMI_FMEQ( VURegs *VU )
{
   if ( _Ft_ == 0 ) return;
}

void recVUMI_FMOR( VURegs *VU )
{
   if ( _Ft_ == 0 ) return;
}

void recVUMI_FCAND( VURegs *VU )
{
   recUpdateFlags( );

   MOV32MtoR( EAX, (u32)&VU->VI[ REG_CLIP_FLAG ].UL );
   AND32ItoR( EAX, VU->code & 0xFFFFFF );
   NEG32R( EAX );
   SBB32RtoR( EAX, EAX );
   NEG32R( EAX );
   MOV32RtoM( (u32)&VU->VI[ 1 ].UL, EAX );
}

void recVUMI_FCEQ( VURegs *VU )
{
}

void recVUMI_FCOR( VURegs *VU )
{
}

void recVUMI_FCSET( VURegs *VU )
{
}

void recVUMI_FCGET( VURegs *VU )
{
}

s32 _recbranchAddr(VURegs * VU) {
	s32 bpc = pc + _Imm11_ * 8; 
	if (bpc < 0) {
		bpc = pc + _UImm11_ * 8; 
	}
	if (VU == &VU1) {
		bpc&= 0x3fff;
	} else {
		bpc&= 0x0fff;
	}

	return bpc;
}

void recVUMI_IBEQ(VURegs *VU) {
	u32 bpc = _recbranchAddr(VU); 

	_freeXMMregs(VU);

	MOV32ItoM( (u32)&VU->VI[ REG_TPC ].UL, (u32)pc );

	MOV16MtoR( EAX, (u32)&VU->VI[ _Ft_ ].US[ 0 ] );
	CMP16MtoR( EAX, (u32)&VU->VI[ _Fs_ ].US[ 0 ] );
	j8Ptr[ 0 ] = JNE8( 0 );
#ifdef __x86_64__
	MOV32ItoR( EDI, (u32)VU );
	CALLFunc( (u32)VU->vuExec );
#else
	PUSH32I( (u32)VU );
	CALLFunc( (u32)VU->vuExec );
	ADD32ItoR( ESP, 4 );
#endif
	MOV32ItoM( (u32)&VU->VI[ REG_TPC ].UL, bpc );

//	RET( ); 

	x86SetJ8( j8Ptr[ 0 ] );
	branch = 1;
}

void recVUMI_IBGEZ( VURegs *VU ) {
   u32 bpc = _recbranchAddr(VU); 

	_freeXMMregs(VU);

	MOV16ItoM( (u32)&VU->VI[ REG_TPC ].UL, (u32)pc );
	CMP16ItoM( (u32)&VU->VI[ _Fs_ ].US[ 0 ], 0x0 );

   j8Ptr[ 0 ] = JL8( 0 );
#ifdef __x86_64__
   MOV32ItoR( EDI, (u32)VU );
   CALLFunc( (u32)VU->vuExec );
#else
   PUSH32I( (u32)VU );
   CALLFunc( (u32)VU->vuExec );
   ADD32ItoR( ESP, 4 );
#endif
	MOV32ItoM( (u32)&VU->VI[ REG_TPC ].UL, bpc );
   //////
//	RET( ); 
   //////
   x86SetJ8( j8Ptr[ 0 ] );
    
   branch = 1;
}

void recVUMI_IBGTZ( VURegs *VU ) {
   u32 bpc = _recbranchAddr(VU); 

	_freeXMMregs(VU);

	MOV16ItoM( (u32)&VU->VI[ REG_TPC ].UL, (u32)pc );
	CMP16ItoM( (u32)&VU->VI[ _Fs_ ].US[ 0 ], 0x0 );

   j8Ptr[ 0 ] = JLE8( 0 );
#ifdef __x86_64__
   MOV32ItoR( EDI, (u32)VU );
   CALLFunc( (u32)VU->vuExec );
#else
   PUSH32I( (u32)VU );
   CALLFunc( (u32)VU->vuExec );
   ADD32ItoR( ESP, 4 );
#endif
	MOV32ItoM( (u32)&VU->VI[ REG_TPC ].UL, bpc );
   //////
//	RET(); 
   //////
   x86SetJ8( j8Ptr[ 0 ] );
    
   branch = 1;
}

void recVUMI_IBLEZ( VURegs *VU ) {
   u32 bpc = _recbranchAddr(VU); 

	_freeXMMregs(VU);

	MOV16ItoM( (u32)&VU->VI[ REG_TPC ].UL, (u32)pc );
	CMP16ItoM( (u32)&VU->VI[ _Fs_ ].US[ 0 ], 0x0 );

   j8Ptr[ 0 ] = JG8( 0 );
#ifdef __x86_64__
   MOV32ItoR( EDI, (u32)VU );
   CALLFunc( (u32)VU->vuExec );
#else
   PUSH32I( (u32)VU );
   CALLFunc( (u32)VU->vuExec );
   ADD32ItoR( ESP, 4 );
#endif
	MOV32ItoM( (u32)&VU->VI[ REG_TPC ].UL, bpc );
   //////
//	RET( ); 
   //////
   x86SetJ8( j8Ptr[ 0 ] );
    
   branch = 1;
}

void recVUMI_IBLTZ( VURegs *VU ) {
   u32 bpc = _recbranchAddr(VU); 

	_freeXMMregs(VU);

	MOV16ItoM( (u32)&VU->VI[ REG_TPC ].UL, (u32)pc );
	CMP16ItoM( (u32)&VU->VI[ _Fs_ ].US[ 0 ], 0x0 );

   j8Ptr[ 0 ] = JGE8( 0 );
#ifdef __x86_64__
   MOV32ItoR( EDI, (u32)VU );
   CALLFunc( (u32)VU->vuExec );
#else
   PUSH32I( (u32)VU );
   CALLFunc( (u32)VU->vuExec );
   ADD32ItoR( ESP, 4 );
#endif
	MOV32ItoM( (u32)&VU->VI[ REG_TPC ].UL, bpc );
   //////
//	RET( ); 
   //////
   x86SetJ8( j8Ptr[ 0 ] );
    
   branch = 1;
}

void recVUMI_IBNE( VURegs *VU ) {
   u32 bpc = _recbranchAddr(VU); 

	_freeXMMregs(VU);

	MOV16ItoM( (u32)&VU->VI[ REG_TPC ].UL, (u32)pc );
	MOV16MtoR( EAX, (u32)&VU->VI[ _Ft_ ].US[ 0 ] );
	CMP16MtoR( EAX, (u32)&VU->VI[ _Fs_ ].US[ 0 ] );

   j8Ptr[ 0 ] = JE8( 0 );
#ifdef __x86_64__
   MOV32ItoR( EDI, (u32)VU );
   CALLFunc( (u32)VU->vuExec );
#else
   PUSH32I( (u32)VU );
   CALLFunc( (u32)VU->vuExec );
   ADD32ItoR( ESP, 4 );
#endif
	MOV32ItoM( (u32)&VU->VI[ REG_TPC ].UL, bpc );
   //////
//	RET( ); 
   //////
   x86SetJ8( j8Ptr[ 0 ] );
    
   branch = 1;
}

void recVUMI_B(VURegs *VU) {
	u32 bpc = _recbranchAddr(VU); 

	_freeXMMregs(VU);

	MOV32ItoM( (u32)&VU->VI[ REG_TPC ].UL, (u32)pc );

#ifdef __x86_64__
   MOV32ItoR( EDI, (u32)VU );
   CALLFunc( (u32)VU->vuExec );
#else
   PUSH32I( (u32)VU );
   CALLFunc( (u32)VU->vuExec );
   ADD32ItoR( ESP, 4 );
#endif
    
	MOV32ItoM( (u32)&VU->VI[ REG_TPC ].UL, bpc );

   branch = 1;
}

void recVUMI_BAL( VURegs *VU ) {
	u32 bpc = _recbranchAddr(VU); 

	_freeXMMregs(VU);

	MOV32ItoM( (u32)&VU->VI[ REG_TPC ].UL, (u32)pc );

	if ( _Ft_ ) {
		MOV16ItoM( (u32)&VU->VI[ _Ft_ ].US[ 0 ], (pc + 8)/8 );
	}

#ifdef __x86_64__
	MOV32ItoR( EDI, (u32)VU );
	CALLFunc( (u32)VU->vuExec );
#else
	PUSH32I( (u32)VU );
	CALLFunc( (u32)VU->vuExec );
	ADD32ItoR( ESP, 4 );
#endif
    
	MOV32ItoM( (u32)&VU->VI[ REG_TPC ].UL, bpc );

	branch = 1;
}

static u16 _branchpc;

void recVUMI_JR(VURegs *VU) {
	_freeXMMregs(VU);

	MOV32ItoM( (u32)&VU->VI[ REG_TPC ].UL, (u32)pc );
	MOV32MtoR(EAX, (u32)&VU->VI[ _Fs_ ].US[ 0 ] );
	SHL32ItoR(EAX, 3);
	MOV16RtoM((u32)&_branchpc, EAX);

#ifdef __x86_64__
	MOV32ItoR( EDI, (u32)VU );
	CALLFunc( (u32)VU->vuExec );
#else
	PUSH32I( (u32)VU );
	CALLFunc( (u32)VU->vuExec );
	ADD32ItoR( ESP, 4 );
#endif
    
	MOV16MtoR(EAX, (u32)&_branchpc);
	MOV16RtoM((u32)&VU->VI[ REG_TPC ].UL, EAX);

	branch = 1;
}

void recVUMI_JALR(VURegs *VU) {
	_freeXMMregs(VU);

	MOV32ItoM( (u32)&VU->VI[ REG_TPC ].UL, (u32)pc );
	MOV32MtoR(EAX, (u32)&VU->VI[ _Fs_ ].US[ 0 ] );
	SHL32ItoR(EAX, 3);
	MOV16RtoM((u32)&_branchpc, EAX);

	if ( _Ft_ ) {
		MOV16ItoM( (u32)&VU->VI[ _Ft_ ].US[ 0 ], (pc + 8)/8 );
	}

#ifdef __x86_64__
	MOV32ItoR( EDI, (u32)VU );
	CALLFunc( (u32)VU->vuExec );
#else
	PUSH32I( (u32)VU );
	CALLFunc( (u32)VU->vuExec );
	ADD32ItoR( ESP, 4 );
#endif
    
	MOV16MtoR(EAX, (u32)&_branchpc);
	MOV16RtoM((u32)&VU->VI[ REG_TPC ].UL, EAX);

	branch = 1;
}

void recVUMI_MFP(VURegs *VU) {
   if ( _Ft_ == 0 ) return;
}

void recVUMI_WAITP(VURegs *VU) {
}

void recVUMI_ESADD( VURegs *VU )
{
}

void recVUMI_ERSADD( VURegs *VU )
{
}

void recVUMI_ELENG( VURegs *VU )
{
}

void recVUMI_ERLENG( VURegs *VU )
{
}

void recVUMI_EATANxy( VURegs *VU )
{
}

void recVUMI_EATANxz( VURegs *VU )
{
}

void recVUMI_ESUM( VURegs *VU )
{
}

void recVUMI_ERCPR( VURegs *VU )
{
}

void recVUMI_ESQRT( VURegs *VU )
{
}

void recVUMI_ERSQRT( VURegs *VU )
{
}

void recVUMI_ESIN( VURegs *VU )
{
}

void recVUMI_EATAN( VURegs *VU )
{
}

void recVUMI_EEXP( VURegs *VU )
{
}

void recVUMI_XITOP( VURegs *VU )
{
}

void recVUMI_XTOP( VURegs *VU )
{
   if ( _Ft_ == 0 ) return;

   MOV32MtoR( EAX, (u32)&VU->vifRegs->top );
   MOV16RtoM( VU_VI_ADDR( _Ft_ ), EAX );
}

